# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class TaskSummary(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'num_total_tasks': 'int',
        'num_finished_tasks': 'int',
        'num_failed_tasks': 'int',
        'num_running_tasks': 'int',
        'num_pending_dependencies_tasks': 'int',
        'num_pending_schedule_tasks': 'int'
    }

    attribute_map = {
        'num_total_tasks': 'num_total_tasks',
        'num_finished_tasks': 'num_finished_tasks',
        'num_failed_tasks': 'num_failed_tasks',
        'num_running_tasks': 'num_running_tasks',
        'num_pending_dependencies_tasks': 'num_pending_dependencies_tasks',
        'num_pending_schedule_tasks': 'num_pending_schedule_tasks'
    }

    def __init__(self, num_total_tasks=None, num_finished_tasks=None, num_failed_tasks=None, num_running_tasks=None, num_pending_dependencies_tasks=None, num_pending_schedule_tasks=None, local_vars_configuration=None):  # noqa: E501
        """TaskSummary - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._num_total_tasks = None
        self._num_finished_tasks = None
        self._num_failed_tasks = None
        self._num_running_tasks = None
        self._num_pending_dependencies_tasks = None
        self._num_pending_schedule_tasks = None
        self.discriminator = None

        self.num_total_tasks = num_total_tasks
        self.num_finished_tasks = num_finished_tasks
        self.num_failed_tasks = num_failed_tasks
        self.num_running_tasks = num_running_tasks
        self.num_pending_dependencies_tasks = num_pending_dependencies_tasks
        self.num_pending_schedule_tasks = num_pending_schedule_tasks

    @property
    def num_total_tasks(self):
        """Gets the num_total_tasks of this TaskSummary.  # noqa: E501


        :return: The num_total_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_total_tasks

    @num_total_tasks.setter
    def num_total_tasks(self, num_total_tasks):
        """Sets the num_total_tasks of this TaskSummary.


        :param num_total_tasks: The num_total_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_total_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_total_tasks`, must not be `None`")  # noqa: E501

        self._num_total_tasks = num_total_tasks

    @property
    def num_finished_tasks(self):
        """Gets the num_finished_tasks of this TaskSummary.  # noqa: E501


        :return: The num_finished_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_finished_tasks

    @num_finished_tasks.setter
    def num_finished_tasks(self, num_finished_tasks):
        """Sets the num_finished_tasks of this TaskSummary.


        :param num_finished_tasks: The num_finished_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_finished_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_finished_tasks`, must not be `None`")  # noqa: E501

        self._num_finished_tasks = num_finished_tasks

    @property
    def num_failed_tasks(self):
        """Gets the num_failed_tasks of this TaskSummary.  # noqa: E501


        :return: The num_failed_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_failed_tasks

    @num_failed_tasks.setter
    def num_failed_tasks(self, num_failed_tasks):
        """Sets the num_failed_tasks of this TaskSummary.


        :param num_failed_tasks: The num_failed_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_failed_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_failed_tasks`, must not be `None`")  # noqa: E501

        self._num_failed_tasks = num_failed_tasks

    @property
    def num_running_tasks(self):
        """Gets the num_running_tasks of this TaskSummary.  # noqa: E501


        :return: The num_running_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_running_tasks

    @num_running_tasks.setter
    def num_running_tasks(self, num_running_tasks):
        """Sets the num_running_tasks of this TaskSummary.


        :param num_running_tasks: The num_running_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_running_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_running_tasks`, must not be `None`")  # noqa: E501

        self._num_running_tasks = num_running_tasks

    @property
    def num_pending_dependencies_tasks(self):
        """Gets the num_pending_dependencies_tasks of this TaskSummary.  # noqa: E501


        :return: The num_pending_dependencies_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_pending_dependencies_tasks

    @num_pending_dependencies_tasks.setter
    def num_pending_dependencies_tasks(self, num_pending_dependencies_tasks):
        """Sets the num_pending_dependencies_tasks of this TaskSummary.


        :param num_pending_dependencies_tasks: The num_pending_dependencies_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_pending_dependencies_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_pending_dependencies_tasks`, must not be `None`")  # noqa: E501

        self._num_pending_dependencies_tasks = num_pending_dependencies_tasks

    @property
    def num_pending_schedule_tasks(self):
        """Gets the num_pending_schedule_tasks of this TaskSummary.  # noqa: E501


        :return: The num_pending_schedule_tasks of this TaskSummary.  # noqa: E501
        :rtype: int
        """
        return self._num_pending_schedule_tasks

    @num_pending_schedule_tasks.setter
    def num_pending_schedule_tasks(self, num_pending_schedule_tasks):
        """Sets the num_pending_schedule_tasks of this TaskSummary.


        :param num_pending_schedule_tasks: The num_pending_schedule_tasks of this TaskSummary.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_pending_schedule_tasks is None:  # noqa: E501
            raise ValueError("Invalid value for `num_pending_schedule_tasks`, must not be `None`")  # noqa: E501

        self._num_pending_schedule_tasks = num_pending_schedule_tasks

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TaskSummary):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, TaskSummary):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ActorTableRow(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'job_id': 'str',
        'class_name': 'str',
        'serialized_runtime_env': 'str',
        'required_resources_serialized': 'str',
        'placement_group_id': 'str',
        'label_selector_serialized': 'str',
        'state': 'ActorState',
        'node_id': 'str',
        'worker_id': 'str',
        'exit_detail_serialized': 'str',
        'transition_timestamp': 'int',
        'ray_session_name': 'str',
        'exit_message': 'str'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'job_id': 'job_id',
        'class_name': 'class_name',
        'serialized_runtime_env': 'serialized_runtime_env',
        'required_resources_serialized': 'required_resources_serialized',
        'placement_group_id': 'placement_group_id',
        'label_selector_serialized': 'label_selector_serialized',
        'state': 'state',
        'node_id': 'node_id',
        'worker_id': 'worker_id',
        'exit_detail_serialized': 'exit_detail_serialized',
        'transition_timestamp': 'transition_timestamp',
        'ray_session_name': 'ray_session_name',
        'exit_message': 'exit_message'
    }

    def __init__(self, id=None, name=None, job_id=None, class_name=None, serialized_runtime_env=None, required_resources_serialized=None, placement_group_id=None, label_selector_serialized=None, state=None, node_id=None, worker_id=None, exit_detail_serialized=None, transition_timestamp=None, ray_session_name=None, exit_message=None, local_vars_configuration=None):  # noqa: E501
        """ActorTableRow - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._job_id = None
        self._class_name = None
        self._serialized_runtime_env = None
        self._required_resources_serialized = None
        self._placement_group_id = None
        self._label_selector_serialized = None
        self._state = None
        self._node_id = None
        self._worker_id = None
        self._exit_detail_serialized = None
        self._transition_timestamp = None
        self._ray_session_name = None
        self._exit_message = None
        self.discriminator = None

        self.id = id
        if name is not None:
            self.name = name
        if job_id is not None:
            self.job_id = job_id
        if class_name is not None:
            self.class_name = class_name
        if serialized_runtime_env is not None:
            self.serialized_runtime_env = serialized_runtime_env
        if required_resources_serialized is not None:
            self.required_resources_serialized = required_resources_serialized
        if placement_group_id is not None:
            self.placement_group_id = placement_group_id
        if label_selector_serialized is not None:
            self.label_selector_serialized = label_selector_serialized
        if state is not None:
            self.state = state
        if node_id is not None:
            self.node_id = node_id
        if worker_id is not None:
            self.worker_id = worker_id
        if exit_detail_serialized is not None:
            self.exit_detail_serialized = exit_detail_serialized
        if transition_timestamp is not None:
            self.transition_timestamp = transition_timestamp
        self.ray_session_name = ray_session_name
        if exit_message is not None:
            self.exit_message = exit_message

    @property
    def id(self):
        """Gets the id of this ActorTableRow.  # noqa: E501


        :return: The id of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ActorTableRow.


        :param id: The id of this ActorTableRow.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this ActorTableRow.  # noqa: E501


        :return: The name of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ActorTableRow.


        :param name: The name of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def job_id(self):
        """Gets the job_id of this ActorTableRow.  # noqa: E501


        :return: The job_id of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this ActorTableRow.


        :param job_id: The job_id of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    @property
    def class_name(self):
        """Gets the class_name of this ActorTableRow.  # noqa: E501


        :return: The class_name of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._class_name

    @class_name.setter
    def class_name(self, class_name):
        """Sets the class_name of this ActorTableRow.


        :param class_name: The class_name of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._class_name = class_name

    @property
    def serialized_runtime_env(self):
        """Gets the serialized_runtime_env of this ActorTableRow.  # noqa: E501


        :return: The serialized_runtime_env of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._serialized_runtime_env

    @serialized_runtime_env.setter
    def serialized_runtime_env(self, serialized_runtime_env):
        """Sets the serialized_runtime_env of this ActorTableRow.


        :param serialized_runtime_env: The serialized_runtime_env of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._serialized_runtime_env = serialized_runtime_env

    @property
    def required_resources_serialized(self):
        """Gets the required_resources_serialized of this ActorTableRow.  # noqa: E501


        :return: The required_resources_serialized of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._required_resources_serialized

    @required_resources_serialized.setter
    def required_resources_serialized(self, required_resources_serialized):
        """Sets the required_resources_serialized of this ActorTableRow.


        :param required_resources_serialized: The required_resources_serialized of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._required_resources_serialized = required_resources_serialized

    @property
    def placement_group_id(self):
        """Gets the placement_group_id of this ActorTableRow.  # noqa: E501


        :return: The placement_group_id of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._placement_group_id

    @placement_group_id.setter
    def placement_group_id(self, placement_group_id):
        """Sets the placement_group_id of this ActorTableRow.


        :param placement_group_id: The placement_group_id of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._placement_group_id = placement_group_id

    @property
    def label_selector_serialized(self):
        """Gets the label_selector_serialized of this ActorTableRow.  # noqa: E501


        :return: The label_selector_serialized of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._label_selector_serialized

    @label_selector_serialized.setter
    def label_selector_serialized(self, label_selector_serialized):
        """Sets the label_selector_serialized of this ActorTableRow.


        :param label_selector_serialized: The label_selector_serialized of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._label_selector_serialized = label_selector_serialized

    @property
    def state(self):
        """Gets the state of this ActorTableRow.  # noqa: E501


        :return: The state of this ActorTableRow.  # noqa: E501
        :rtype: ActorState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this ActorTableRow.


        :param state: The state of this ActorTableRow.  # noqa: E501
        :type: ActorState
        """

        self._state = state

    @property
    def node_id(self):
        """Gets the node_id of this ActorTableRow.  # noqa: E501


        :return: The node_id of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._node_id

    @node_id.setter
    def node_id(self, node_id):
        """Sets the node_id of this ActorTableRow.


        :param node_id: The node_id of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._node_id = node_id

    @property
    def worker_id(self):
        """Gets the worker_id of this ActorTableRow.  # noqa: E501


        :return: The worker_id of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._worker_id

    @worker_id.setter
    def worker_id(self, worker_id):
        """Sets the worker_id of this ActorTableRow.


        :param worker_id: The worker_id of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._worker_id = worker_id

    @property
    def exit_detail_serialized(self):
        """Gets the exit_detail_serialized of this ActorTableRow.  # noqa: E501


        :return: The exit_detail_serialized of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._exit_detail_serialized

    @exit_detail_serialized.setter
    def exit_detail_serialized(self, exit_detail_serialized):
        """Sets the exit_detail_serialized of this ActorTableRow.


        :param exit_detail_serialized: The exit_detail_serialized of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._exit_detail_serialized = exit_detail_serialized

    @property
    def transition_timestamp(self):
        """Gets the transition_timestamp of this ActorTableRow.  # noqa: E501


        :return: The transition_timestamp of this ActorTableRow.  # noqa: E501
        :rtype: int
        """
        return self._transition_timestamp

    @transition_timestamp.setter
    def transition_timestamp(self, transition_timestamp):
        """Sets the transition_timestamp of this ActorTableRow.


        :param transition_timestamp: The transition_timestamp of this ActorTableRow.  # noqa: E501
        :type: int
        """

        self._transition_timestamp = transition_timestamp

    @property
    def ray_session_name(self):
        """Gets the ray_session_name of this ActorTableRow.  # noqa: E501


        :return: The ray_session_name of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._ray_session_name

    @ray_session_name.setter
    def ray_session_name(self, ray_session_name):
        """Sets the ray_session_name of this ActorTableRow.


        :param ray_session_name: The ray_session_name of this ActorTableRow.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_session_name is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_session_name`, must not be `None`")  # noqa: E501

        self._ray_session_name = ray_session_name

    @property
    def exit_message(self):
        """Gets the exit_message of this ActorTableRow.  # noqa: E501


        :return: The exit_message of this ActorTableRow.  # noqa: E501
        :rtype: str
        """
        return self._exit_message

    @exit_message.setter
    def exit_message(self, exit_message):
        """Sets the exit_message of this ActorTableRow.


        :param exit_message: The exit_message of this ActorTableRow.  # noqa: E501
        :type: str
        """

        self._exit_message = exit_message

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ActorTableRow):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ActorTableRow):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass
from enum import Enum
import importlib.resources
from io import StringIO
import json
import pathlib
import shlex
import subprocess
import sys
import tempfile
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

import click
from rich.console import Console
from rich.table import Table
import yaml

from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.exceptions import (
    ApiException,
    ApiTypeError,
    ApiValueError,
)
from anyscale.commands import command_examples
from anyscale.commands.list_util import (
    display_list,
    MAX_PAGE_SIZE,
    NON_INTERACTIVE_DEFAULT_MAX_ITEMS,
    validate_page_size,
)
from anyscale.commands.util import (
    AnyscaleCommand,
    build_kv_table,
    convert_kv_strings_to_dict,
    parse_repeatable_tags_to_dict,
    parse_tags_kv_to_str_map,
)
from anyscale.util import (
    AnyscaleJSONEncoder,
    get_endpoint,
    validate_non_negative_arg,
    validate_workspace_state_filter,
)
import anyscale.workspace
from anyscale.workspace._private.workspace_sdk import ANYSCALE_WORKSPACES_SSH_OPTIONS
from anyscale.workspace.commands import _WORKSPACE_SDK_SINGLETON_KEY
from anyscale.workspace.models import (
    UpdateWorkspaceConfig,
    Workspace,
    WorkspaceConfig,
    WorkspaceSortField,
    WorkspaceSortOrder,
    WorkspaceState,
)


log = BlockLogger()  # CLI Logger

# Constants for SSH configuration
HTTPS_PORT = "443"
SSH_TEST_TIMEOUT_SECONDS = 8
WSS_PATH = "/sshws"
PREFERRED_AUTH_METHOD = "PreferredAuthentications=publickey"


def validate_max_items(ctx, param, value):
    """Click callback to validate max_items argument."""
    if value is None:
        return None
    return validate_non_negative_arg(ctx, param, value)


def _parse_sort_option(
    sort: Optional[str],
) -> Tuple[Optional[str], WorkspaceSortOrder]:
    """Parse sort string (e.g. "-created_at") into field and order."""
    if not sort:
        return None, WorkspaceSortOrder.ASC

    # Build case-insensitive map of allowed fields
    allowed = {
        f.value.lower(): f.value for f in WorkspaceSortField.__members__.values()
    }

    # Detect leading '-' for descending
    if sort.startswith("-"):
        raw = sort[1:]
        order = WorkspaceSortOrder.DESC
    else:
        raw = sort
        order = WorkspaceSortOrder.ASC

    key = raw.lower()
    if key not in allowed:
        allowed_names = ", ".join(sorted(allowed.values()))
        raise click.BadParameter(
            f"Invalid sort field '{raw}'. Allowed: {allowed_names}"
        )

    return allowed[key], order


def _create_workspace_list_table(show_header: bool) -> Table:
    table = Table(show_header=show_header, expand=True)
    # Allow wrapping for all columns to prevent text cutoff
    table.add_column("NAME", no_wrap=False, overflow="fold", ratio=3, min_width=15)
    table.add_column("ID", no_wrap=False, overflow="fold", ratio=2, min_width=20)
    table.add_column("STATE", no_wrap=False, overflow="fold", ratio=2, min_width=12)
    table.add_column("PROJECT", no_wrap=False, overflow="fold", ratio=2, min_width=15)
    table.add_column("CLOUD", no_wrap=False, overflow="fold", ratio=2, min_width=12)
    table.add_column(
        "CREATED BY", no_wrap=False, overflow="fold", ratio=3, min_width=24
    )
    table.add_column(
        "CREATED AT", no_wrap=False, overflow="fold", ratio=2, min_width=20
    )
    return table


def _create_workspace_list_table_verbose(show_header: bool) -> Table:
    table = Table(show_header=show_header, expand=True, box=None)
    table.add_column("NAME", overflow="fold", no_wrap=False)
    table.add_column("ID", overflow="fold", no_wrap=False)
    table.add_column("STATE", overflow="fold", no_wrap=False)
    table.add_column("PROJECT", overflow="fold", no_wrap=False)
    table.add_column("CLOUD", overflow="fold", no_wrap=False)
    table.add_column("CLUSTER", overflow="fold", no_wrap=False)
    table.add_column("CREATED BY", overflow="fold", no_wrap=False)
    table.add_column("CREATED AT", overflow="fold", no_wrap=False)
    table.add_column("LAST STARTED AT", overflow="fold", no_wrap=False)
    return table


def _format_workspace_output(workspace: Workspace) -> Dict[str, str]:
    created_at = (
        workspace.created_at.strftime("%Y-%m-%d %H:%M") if workspace.created_at else ""
    )

    return {
        "name": workspace.name,
        "id": workspace.id,
        "state": str(workspace.state),
        "project": workspace.project_id or "",
        "cloud": workspace.cloud_id or "",
        "created_by": workspace.creator_email or workspace.creator_id or "",
        "created_at": created_at,
    }


def _format_workspace_output_verbose(workspace: Workspace) -> Dict[str, str]:
    created_at = (
        workspace.created_at.strftime("%Y-%m-%d %H:%M") if workspace.created_at else ""
    )
    last_started = (
        workspace.last_started_at.strftime("%Y-%m-%d %H:%M")
        if workspace.last_started_at
        else ""
    )

    return {
        "name": workspace.name,
        "id": workspace.id,
        "state": str(workspace.state),
        "project": workspace.project_id or "",
        "cloud": workspace.cloud_id or "",
        "cluster": workspace.cluster_id or "",
        "created_by": workspace.creator_email or workspace.creator_id or "",
        "created_at": created_at,
        "last_started_at": last_started,
    }


def _validate_workspace_name_and_id(
    name: Optional[str], id: Optional[str]  # noqa: A002
):
    if name is None and id is None:
        raise click.ClickException("One of '--name' and '--id' must be provided.")

    if name is not None and id is not None:
        raise click.ClickException("Only one of '--name' and '--id' can be provided.")


def _check_workspace_is_running(
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Verify that the workspace is in RUNNING state."""
    try:
        workspace_status = anyscale.workspace.status(
            name=name, id=id, cloud=cloud, project=project
        )
        if workspace_status != WorkspaceState.RUNNING:
            raise click.ClickException(
                f"Workspace must be running to SSH into it. Current status: {workspace_status}"
            )
    except ValueError as e:
        # Handle workspace not found or other value errors
        error_msg = str(e)
        if "not found" in error_msg.lower():
            workspace_identifier = name if name else id
            raise click.ClickException(
                f"Workspace '{workspace_identifier}' not found. Please check the workspace name/id and try again."
            )
        else:
            raise click.ClickException(f"Error checking workspace status: {error_msg}")
    except (AttributeError, KeyError, TypeError):
        # Handle any other errors from status check
        raise click.ClickException(
            "Failed to check workspace status. Please ensure the workspace exists and you have access to it."
        )


def _get_workspace_directory_name(
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    cloud: Optional[str],
    project: Optional[str],
) -> str:
    """Get the default directory name for the workspace."""
    try:
        workspace_private_sdk = _LAZY_SDK_SINGLETONS[_WORKSPACE_SDK_SINGLETON_KEY]
        return workspace_private_sdk.get_default_dir_name(
            name=name, id=id, cloud=cloud, project=project
        )
    except (AttributeError, KeyError, ValueError, TypeError):
        # Handle errors getting default directory name
        raise click.ClickException(
            "Failed to retrieve workspace configuration. Please try again later."
        )


def _create_directory_setup_command(dir_name: str) -> str:
    """Create shell command to set up the workspace directory."""
    # Even though dir_name comes from our API, we should still shell escape it.
    dir_name_escaped = shlex.quote(dir_name)
    return (
        f'if [ -d "$HOME/{dir_name_escaped}" ]; then '
        f'cd "$HOME/{dir_name_escaped}"; '
        f"else "
        f'mkdir -p "$HOME/{dir_name_escaped}" 2>/dev/null && '
        f'cd "$HOME/{dir_name_escaped}" && '
        f'echo "Created directory $HOME/{dir_name_escaped} (it did not exist)." >&2 || '
        f'echo "Warning: Could not access or create directory $HOME/{dir_name_escaped}. Staying in home directory." >&2; '
        f"fi"
    )


class ConnectionType(Enum):
    HTTPS = "https"
    LEGACY = "legacy"


@dataclass
class SSHConfig:
    """Configuration for SSH connection to workspace."""

    target_host: str
    config_file: str
    connection_type: ConnectionType
    proxy_command: Optional[str] = None
    port: Optional[str] = None

    @property
    def ssh_options(self) -> List[str]:
        """Get SSH options based on connection type."""
        if self.connection_type == ConnectionType.HTTPS:
            options = [
                "-p",
                self.port or HTTPS_PORT,
                "-o",
                PREFERRED_AUTH_METHOD,
            ]
            if self.proxy_command:
                options.extend(["-o", f"ProxyCommand={self.proxy_command}"])
            return options
        return []


def _setup_https_connection(
    workspace_obj: Workspace, workspace_private_sdk, host_name: str, config_file: str
) -> SSHConfig:
    """Set up HTTPS connection and return SSHConfig object."""
    cluster = workspace_private_sdk.client.get_workspace_cluster(workspace_obj.id)

    if not cluster:
        raise click.ClickException(
            "Could not retrieve cluster details for the workspace."
        )

    # Get hostname with multiple fallback methods
    public_hostname = _get_public_hostname(cluster)

    # Get cluster access token
    cluster_access_token = _get_cluster_access_token(cluster, workspace_private_sdk)

    # Set up proxy command
    proxy_cmd = _create_proxy_command(public_hostname, cluster_access_token)

    return SSHConfig(
        target_host=f"ray@{host_name}",
        config_file=config_file,
        connection_type=ConnectionType.HTTPS,
        proxy_command=proxy_cmd,
        port=HTTPS_PORT,
    )


def _get_public_hostname(cluster) -> str:
    """Extract public hostname from cluster with fallback methods."""
    # Attempt 1: Use cluster.hostname
    if hasattr(cluster, "hostname") and cluster.hostname:
        return cluster.hostname

    # Attempt 2: Fallback to parsing webterminal_auth_url
    if hasattr(cluster, "webterminal_auth_url") and cluster.webterminal_auth_url:
        parsed_url = urlparse(cluster.webterminal_auth_url)
        if parsed_url.netloc:
            return parsed_url.netloc

        # webterminal_auth_url was present but parsing failed
        raise click.ClickException(
            "Could not extract hostname from cluster configuration. "
            "The URL appears to be malformed."
        )

    # Both methods failed
    raise click.ClickException(
        "Could not retrieve hostname for HTTPS connection. "
        "Required cluster configuration is not available."
    )


def _get_cluster_access_token(cluster, workspace_private_sdk) -> str:
    """Get cluster access token for HTTPS connection."""
    if not hasattr(cluster, "id") or not cluster.id:
        raise click.ClickException(
            "Cluster configuration is incomplete, cannot retrieve access token."
        )

    try:
        # We need to use the internal API here as there's no public API available
        # This might be updated in future SDK versions
        cluster_access_token = workspace_private_sdk.client._internal_api_client.get_cluster_access_token_api_v2_authentication_cluster_id_cluster_access_token_get(  # noqa: SLF001
            cluster_id=cluster.id
        )

        if not cluster_access_token:
            raise click.ClickException(
                "Failed to retrieve authentication token. Please try again."
            )

        return cluster_access_token

    except (ApiException, ApiTypeError, ApiValueError):
        # Don't expose API details in error messages
        raise click.ClickException(
            "Failed to authenticate. Please check your permissions and try again."
        ) from None
    except click.ClickException:
        raise  # Re-raise click exceptions as-is
    except (AttributeError, KeyError, ValueError, TypeError, RuntimeError) as e:
        # Generic error without exposing internal details
        raise click.ClickException(
            "An error occurred during authentication. Please try again."
        ) from e


def _create_proxy_command(public_hostname: str, cluster_access_token: str) -> str:
    """Create the proxy command for HTTPS connection."""
    wss_url = f"wss://{public_hostname}{WSS_PATH}"

    try:
        with importlib.resources.path(
            "anyscale.utils", "ssh_websocket_proxy.py"
        ) as proxy_path:
            proxy_script_path = str(proxy_path)
    except (ModuleNotFoundError, ImportError) as e:
        raise click.ClickException(f"Could not locate SSH proxy script: {e}") from e

    # Properly escape the proxy command arguments
    return " ".join(
        [
            shlex.quote(sys.executable),
            shlex.quote(proxy_script_path),
            shlex.quote(wss_url),
            shlex.quote(cluster_access_token),
        ]
    )


def _build_ssh_command(
    ssh_config: SSHConfig, user_args: List[str], shell_command: str,
) -> List[str]:
    """Build the final SSH command with all options."""
    # Build SSH command with basic options
    base_cmd = (
        ["ssh"]
        + ANYSCALE_WORKSPACES_SSH_OPTIONS
        + ssh_config.ssh_options
        + [ssh_config.target_host, "-F", ssh_config.config_file]
    )

    # Process user-supplied arguments
    if not user_args:
        # No user args, use default interactive shell
        return base_cmd + ["-tt", f"bash -c '{shell_command} && exec bash -i'"]

    # Parse user arguments into options and commands
    user_options, user_command = _parse_user_args(user_args)

    ssh_cmd = base_cmd + user_options

    if user_command:
        # User supplied their own command, use it directly
        ssh_cmd.extend(user_command)
    else:
        # Only options provided, add interactive shell
        # Use -tt for interactive shell unless -T or -N was specified
        if not any(opt in {"-T", "-N"} for opt in user_options):
            ssh_cmd.append("-tt")
        ssh_cmd.append(f"bash -c '{shell_command} && exec bash -i'")

    return ssh_cmd


def _parse_user_args(user_args: List[str]) -> Tuple[List[str], List[str]]:
    """Parse user arguments into options and commands."""
    # Find where command section starts (first non-option argument)
    command_start_idx = None
    for i, arg in enumerate(user_args):
        if arg and not arg.startswith("-"):
            command_start_idx = i
            break

    if command_start_idx is not None:
        user_options = [arg for arg in user_args[:command_start_idx] if arg]
        user_command = [arg for arg in user_args[command_start_idx:] if arg]
    else:
        user_options = [arg for arg in user_args if arg]
        user_command = []

    return user_options, user_command


def _test_https_connectivity(
    workspace_obj: Workspace, workspace_private_sdk, host_name: str, config_file: str,
) -> bool:
    """Test HTTPS SSH connectivity with a quick command. Returns True if available."""
    try:
        cluster = workspace_private_sdk.client.get_workspace_cluster(workspace_obj.id)
        if not cluster:
            return False

        ssh_config = _setup_https_connection(
            workspace_obj, workspace_private_sdk, host_name, config_file
        )

        # Build a test command using the same logic as the actual connection
        # but with a simple echo command that exits immediately
        test_args = [
            "-o",
            "ConnectTimeout=5",
            "-o",
            "BatchMode=yes",
            "echo",
            "connectivity_test",
        ]
        test_cmd = _build_ssh_command(ssh_config, test_args, "")

        # Run test with a timeout
        result = subprocess.run(
            test_cmd,
            check=False,
            timeout=SSH_TEST_TIMEOUT_SECONDS,
            capture_output=True,
            text=True,
        )

        # Check if we got the expected output
        # Connection failed - no need to show error code to user
        return result.returncode == 0 and "connectivity_test" in result.stdout

    except subprocess.TimeoutExpired:
        # Silent failure - the main code will show a user-friendly message
        return False
    except OSError:
        # Silent failure
        return False
    except (click.ClickException, ValueError, AttributeError, KeyError, TypeError):
        # Silent failure
        return False


def _execute_https_ssh(
    workspace_obj: Workspace,
    workspace_private_sdk,
    host_name: str,
    config_file: str,
    ctx_args: List[str],
    shell_command: str,
) -> None:
    """Execute HTTPS SSH connection without timeout."""
    ssh_config = _setup_https_connection(
        workspace_obj, workspace_private_sdk, host_name, config_file
    )
    ssh_cmd = _build_ssh_command(ssh_config, ctx_args, shell_command)
    # Run the actual SSH session without any timeout
    subprocess.run(ssh_cmd, check=False)


def _execute_legacy_ssh(
    ssh_target_host: str, config_file: str, ctx_args: List[str], shell_command: str,
) -> None:
    """Execute legacy SSH connection."""
    legacy_ssh_config = SSHConfig(
        target_host=ssh_target_host,
        config_file=config_file,
        connection_type=ConnectionType.LEGACY,
        proxy_command=None,
        port=None,
    )

    ssh_cmd = _build_ssh_command(legacy_ssh_config, ctx_args, shell_command)
    subprocess.run(ssh_cmd, check=False)


@click.group("workspace_v2", help="Anyscale workspace commands V2.")
def workspace_cli() -> None:
    pass


@workspace_cli.command(
    name="create",
    help="Create a workspace on Anyscale.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_CREATE_EXAMPLE,
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to deploy. When deploying from a file, import path and arguments cannot be provided. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "-n", "--name", required=False, help="Name of the workspace to create.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the workspace. This is exclusive with --containerfile.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the workspace. This is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If you don't specify a Ray version, Anyscale defaults to the latest Ray version available at the time of the Anyscale CLI/SDK release.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the workspace.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the workspace. These will be installed on top of the image.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the workspace. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
def create(  # noqa: PLR0913, PLR0912, C901
    config_file: Optional[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    requirements: Optional[str],
    env: Optional[Tuple[str]],
    tags: Optional[Tuple[str]],
) -> None:
    """Creates a new workspace.

    A name must be provided, either in the file or in the arguments.

    `$ anyscale workspace_v2 create -n my-workspace`

    or add all the information in the config file and do:

    `$ anyscale workspace_v2 create -f config-file.yaml`

    Command-line flags override values in the config file.
    """
    if config_file is not None:
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = WorkspaceConfig.from_yaml(config_file)
    else:
        config = WorkspaceConfig()

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if not config.name:
        raise click.ClickException("Workspace name must be configured")

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if cloud is not None:
        config = config.options(cloud=cloud)
    if project is not None:
        config = config.options(project=project)

    if requirements is not None:
        if not pathlib.Path(requirements).is_file():
            raise click.ClickException(f"Requirements file '{requirements}' not found.")
        config = config.options(requirements=requirements)
    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    if tags:
        tag_map = parse_tags_kv_to_str_map(tags)
        if tag_map:
            config = config.options(tags=tag_map)

    anyscale.workspace.create(config,)


@workspace_cli.command(
    name="start",
    short_help="Starts a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_START_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def start(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Start a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.start(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(
    name="terminate",
    short_help="Terminate a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_TERMINATE_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def terminate(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Terminate a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.terminate(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(
    name="status",
    short_help="Get the status of a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_STATUS_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def status(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Get the status of a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    status = anyscale.workspace.status(name=name, id=id, cloud=cloud, project=project)
    log.info(status)


@workspace_cli.command(
    name="wait",
    short_help="Wait for a workspace to reach a certain status.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_WAIT_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--timeout-s",
    required=False,
    default=1800,
    type=float,
    help="The maximum time in seconds to wait for the workspace to reach the desired state. Default to 30 minutes.",
)
@click.option(
    "--state",
    required=False,
    default=WorkspaceState.RUNNING,
    type=str,
    help="The desired terminal state to wait for. Default is 'RUNNING'.",
)
def wait(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    timeout_s: float,
    state: str,
) -> None:
    """Wait for a workspace to reach a terminal state.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    try:
        state = WorkspaceState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))

    try:
        anyscale.workspace.wait(
            name=name,
            id=id,
            cloud=cloud,
            project=project,
            timeout_s=timeout_s,
            state=state,
        )
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(str(e)) from None


@workspace_cli.group("tags", help="Manage tags for workspaces.")
def workspace_tags_cli() -> None:
    pass


@workspace_tags_cli.command(
    name="add",
    help="Add or update tags on a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_TAGS_ADD_EXAMPLE,  # type: ignore[attr-defined]
)
@click.option(
    "--id", "workspace_id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="Cloud name (used when resolving by name).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Project name (used when resolving by name).",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
def add_tags(
    workspace_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    tags: Tuple[str],
) -> None:
    if not workspace_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = parse_tags_kv_to_str_map(tags)
    if not tag_map:
        raise click.ClickException("Provide at least one --tag key=value.")
    anyscale.workspace.add_tags(  # type: ignore
        id=workspace_id, name=name, cloud=cloud, project=project, tags=tag_map
    )
    stderr = Console(stderr=True)
    ident = workspace_id or name or "<unknown>"
    stderr.print(f"Tags updated for workspace '{ident}'.")


@workspace_tags_cli.command(
    name="remove",
    help="Remove tags by key from a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_TAGS_REMOVE_EXAMPLE,  # type: ignore[attr-defined]
)
@click.option(
    "--id", "workspace_id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="Cloud name (used when resolving by name).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Project name (used when resolving by name).",
)
@click.option("--key", "keys", multiple=True, help="Tag key to remove. Repeatable.")
def remove_tags(
    workspace_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    keys: Tuple[str],
) -> None:
    if not workspace_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    key_list = [k for k in keys if k and k.strip()]
    if not key_list:
        raise click.ClickException("Provide at least one --key to remove.")
    anyscale.workspace.remove_tags(  # type: ignore
        id=workspace_id, name=name, cloud=cloud, project=project, keys=key_list
    )
    stderr = Console(stderr=True)
    ident = workspace_id or name or "<unknown>"
    stderr.print(f"Removed tag keys {key_list} from workspace '{ident}'.")


@workspace_tags_cli.command(
    name="list",
    help="List tags for a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_TAGS_LIST_EXAMPLE,  # type: ignore[attr-defined]
)
@click.option(
    "--id", "workspace_id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="Cloud name (used when resolving by name).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Project name (used when resolving by name).",
)
@click.option("--json", "json_output", is_flag=True, default=False, help="JSON output.")
def list_tags(
    workspace_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    json_output: bool,
) -> None:
    if not workspace_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = anyscale.workspace.list_tags(  # type: ignore
        id=workspace_id, name=name, cloud=cloud, project=project
    )
    if json_output:
        Console().print_json(json=json.dumps(tag_map, indent=2))
    else:
        stderr = Console(stderr=True)
        if not tag_map:
            stderr.print("No tags found.")
            return
        pairs = tag_map.items()
        stderr.print(build_kv_table(pairs, title="Tags"))


@workspace_cli.command(
    name="ssh",
    short_help="SSH into a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_SSH_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--legacy",
    is_flag=True,
    default=False,
    help="Use legacy SSH connection method, bypassing HTTPS SSH.",
)
@click.pass_context
def ssh(
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    legacy: bool,
) -> None:
    """SSH into a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    You may pass extra args for the ssh command, for example to setup port forwarding:
    anyscale workspace_v2 ssh -n workspace-name -- -L 9000:localhost:9000

    Use the --legacy flag to bypass HTTPS SSH and use the legacy connection method directly.
    """
    try:
        _validate_workspace_name_and_id(name=name, id=id)

        # Verify workspace is running
        _check_workspace_is_running(name, id, cloud, project)

        # Inform user that connection might take time (earliest point after verifying workspace is running)
        connection_mode = " (Legacy SSH)" if legacy else ""
        print(
            f"Connecting to workspace{connection_mode}... This might take a while. Press Ctrl+C to cancel."
        )

        # Get workspace directory name
        dir_name = _get_workspace_directory_name(name, id, cloud, project)

        # Create the shell command that will:
        # 1. Try to cd to the directory
        # 2. If that fails, create it and cd to it
        # 3. If that fails too, just stay in home directory with a warning
        shell_command = _create_directory_setup_command(dir_name)

        with tempfile.TemporaryDirectory() as tmpdirname:
            try:
                host_name, config_file = anyscale.workspace.generate_ssh_config_file(
                    name=name,
                    id=id,
                    cloud=cloud,
                    project=project,
                    ssh_config_path=tmpdirname,
                )
            except ValueError as e:
                error_msg = str(e)
                if "not found" in error_msg.lower():
                    workspace_identifier = name if name else id
                    raise click.ClickException(
                        f"Workspace '{workspace_identifier}' not found or not accessible."
                    )
                else:
                    raise click.ClickException("Failed to generate SSH configuration.")
            except (
                OSError,
                IOError,
                RuntimeError,
                AttributeError,
                KeyError,
                TypeError,
            ):
                # Handle any other errors from SSH config generation
                raise click.ClickException(
                    "Failed to generate SSH configuration. Please check your network connection and try again."
                )

            # Get workspace and cluster information to determine connection method
            ssh_target_host = host_name

            # Skip HTTPS if --legacy flag is used
            https_connection_successful = False
            if not legacy:
                # Try HTTPS first (unless legacy flag is set)
                try:
                    workspace_obj = anyscale.workspace.get(
                        name=name, id=id, cloud=cloud, project=project
                    )
                    workspace_private_sdk = _LAZY_SDK_SINGLETONS[
                        _WORKSPACE_SDK_SINGLETON_KEY
                    ]
                    cluster = workspace_private_sdk.client.get_workspace_cluster(
                        workspace_obj.id
                    )

                    if cluster:
                        https_connection_successful = _test_https_connectivity(
                            workspace_obj,
                            workspace_private_sdk,
                            host_name,
                            config_file,
                        )

                except (ValueError, AttributeError, KeyError, TypeError):
                    # If we can't get workspace/cluster info, proceed with legacy SSH
                    pass

            # Execute the appropriate SSH connection based on test results
            if https_connection_successful:
                # HTTPS connectivity test passed, run actual SSH session without timeout
                _execute_https_ssh(
                    workspace_obj,
                    workspace_private_sdk,
                    host_name,
                    config_file,
                    ctx.args,
                    shell_command,
                )
            else:
                # HTTPS test failed or --legacy was specified, use legacy SSH
                if not legacy:  # Only show message if we tried HTTPS first
                    print("Connecting via standard SSH...")
                _execute_legacy_ssh(
                    ssh_target_host, config_file, ctx.args, shell_command
                )

    except click.ClickException:
        # Re-raise click exceptions as they already have user-friendly messages
        raise
    except KeyboardInterrupt:
        # Handle Ctrl+C gracefully
        raise click.ClickException("SSH connection cancelled by user.")
    except (
        OSError,
        IOError,
        RuntimeError,
        ValueError,
        AttributeError,
        KeyError,
        TypeError,
    ):
        # Catch any unexpected exceptions and provide a generic user-friendly message
        raise click.ClickException(
            "An unexpected error occurred while establishing SSH connection. "
            "Please try again or contact support if the issue persists."
        )


@workspace_cli.command(
    name="run_command",
    short_help="Run a command in a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_RUN_COMMAND_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.argument("command", type=str)
def run_command(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    command: str,
) -> None:
    """Run a command in a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.run_command(
        name=name, id=id, cloud=cloud, project=project, command=command
    )


@workspace_cli.command(
    name="pull",
    short_help="Pull the working directory of a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_PULL_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--local-dir",
    required=False,
    default=None,
    type=str,
    help="Local directory to pull the workspace directory to. If not provided, the current directory will be used.",
)
@click.option(
    "--pull-git-state",
    required=False,
    default=False,
    is_flag=True,
    help="Pull the git state of the workspace.",
)
@click.option(
    "--delete",
    required=False,
    default=False,
    is_flag=True,
    help="Delete files in the local directory that are not in the workspace. Excluded files are preserved.",
)
@click.pass_context
def pull(  # noqa: PLR0913
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    local_dir: Optional[str],
    pull_git_state: bool = False,
    delete: bool = False,
) -> None:
    """Pull the working directory of a workspace. New files will be created, existing files will be overwritten.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    This command depends on rsync, please make sure it is installed on your system.

    The --delete flag removes files in the local directory that don't exist in the workspace.
    Excluded files (like .git unless --pull-git-state is used) are preserved and not deleted.

    You may pass extra args for the rsync command, for example to exclude files:
    anyscale workspace_v2 pull -n workspace-name -- --exclude='log.txt'
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.pull(
        name=name,
        id=id,
        cloud=cloud,
        project=project,
        local_dir=local_dir,
        pull_git_state=pull_git_state,
        rsync_args=ctx.args,
        delete=delete,
    )


@workspace_cli.command(
    name="push",
    short_help="Push a local directory to a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_PUSH_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--local-dir",
    required=False,
    default=None,
    type=str,
    help="Local directory to push to the workspace. If not provided, the current directory will be used.",
)
@click.option(
    "--push-git-state",
    required=False,
    default=False,
    is_flag=True,
    help="Push the git state of the workspace.",
)
@click.option(
    "--delete",
    required=False,
    default=False,
    is_flag=True,
    help="Delete files in the workspace that are not in the local directory. Excluded files are preserved.",
)
@click.pass_context
def push(  # noqa: PLR0913
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    local_dir: Optional[str],
    push_git_state: bool = False,
    delete: bool = False,
) -> None:
    """Push a local directory to a workspace. New files will be created, existing files will be overwritten.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    This command depends on rsync, please make sure it is installed on your system.

    The --delete flag removes files in the workspace that don't exist locally.
    Excluded files (like .git unless --push-git-state is used) are preserved and not deleted.

    You may pass extra args for the rsync command, for example to exclude files:
    anyscale workspace_v2 push -n workspace-name -- --exclude='log.txt'
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.push(
        name=name,
        id=id,
        cloud=cloud,
        project=project,
        local_dir=local_dir,
        push_git_state=push_git_state,
        rsync_args=ctx.args,
        delete=delete,
    )


@workspace_cli.command(
    name="update",
    help="Update an existing workspace on Anyscale.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_UPDATE_EXAMPLE,
)
@click.argument("workspace-id", type=str)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to update. Command-line flags will overwrite values read from the file. Unspecified fields will retain their current values, while specified fields will be updated.",
)
@click.option(
    "-n", "--name", required=False, help="New name of the workspace.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="New container image to use for the workspace. This is exclusive with --containerfile.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the workspace. This is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="New Ray version (X.Y.Z) to use with the image specified by --image-uri. This is only used when --image-uri is provided. If not provided, the latest Ray version will be used.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="New named compute configuration to use for the workspace.",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the workspace. These will be installed on top of the image.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="New environment variables to set for the workspace. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
def update(  # noqa: PLR0913, PLR0912
    workspace_id: str,
    config_file: Optional[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    requirements: Optional[str],
    env: Optional[Tuple[str]],
) -> None:
    """Updates an existing workspace.

    Example:
    `$ anyscale workspace_v2 update <workspace-id> --name new-name`

    Command-line flags override values in the config file.
    Unspecified fields will retain their current values, while specified fields will be updated.
    """

    if config_file is not None:
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = UpdateWorkspaceConfig.from_yaml(config_file)
    else:
        config = UpdateWorkspaceConfig()

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if requirements is not None:
        if not pathlib.Path(requirements).is_file():
            raise click.ClickException(f"Requirements file '{requirements}' not found.")
        config = config.options(requirements=requirements)

    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    # Apply the update
    anyscale.workspace.update(id=workspace_id, config=config)


@workspace_cli.command(
    name="get",
    short_help="Get a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_GET_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "-j",
    "--json",
    "json_output",
    is_flag=True,
    default=False,
    help="Output the workspace in a structured JSON format.",
)
@click.option(
    "--yaml", "yaml_output", is_flag=True, default=False, help="Output as YAML.",
)
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Include verbose details.",
)
def get(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    json_output: bool,
    yaml_output: bool,
    verbose: bool,
) -> None:
    """Retrieve workspace details by name or ID.

    Use --name to specify by name or --id for the workspace ID; using both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)

    try:
        # For JSON/YAML, include full config; for table output, skip expensive config fetch
        include_config = json_output or yaml_output

        workspace: Workspace = anyscale.workspace.get(
            name=name,
            id=id,
            cloud=cloud,
            project=project,
            include_config=include_config,
        )

        if json_output:
            print(json.dumps(workspace.to_dict(), indent=2, cls=AnyscaleJSONEncoder))
        elif yaml_output:
            stream = StringIO()
            yaml.safe_dump(workspace.to_dict(), stream, sort_keys=False)
            print(stream.getvalue(), end="")
        else:
            # Use Console() with force_terminal=False for testability
            console = Console(force_terminal=False)

            if verbose:
                table = _create_workspace_list_table_verbose(show_header=True)
                formatted = _format_workspace_output_verbose(workspace)
                table.add_row(
                    formatted["name"],
                    formatted["id"],
                    formatted["state"],
                    formatted["project"],
                    formatted["cloud"],
                    formatted["cluster"],
                    formatted["created_by"],
                    formatted["created_at"],
                    formatted["last_started_at"],
                )
            else:
                table = _create_workspace_list_table(show_header=True)
                formatted = _format_workspace_output(workspace)
                table.add_row(
                    formatted["name"],
                    formatted["id"],
                    formatted["state"],
                    formatted["project"],
                    formatted["cloud"],
                    formatted["created_by"],
                    formatted["created_at"],
                )

            console.print(table)

    except ValueError as e:
        raise click.ClickException(str(e)) from None
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(f"Failed to get workspace: {e}") from None


@workspace_cli.command(
    name="list",
    help="List workspaces.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_LIST_EXAMPLE,
)
@click.option("--workspace-id", "--id", help="ID of the workspace to display.")
@click.option("--name", "-n", help="Substring to match against the workspace name.")
@click.option("--project", help="Filter workspaces by project name.")
@click.option("--cloud", help="Filter workspaces by cloud name.")
@click.option(
    "--created-by-me",
    is_flag=True,
    default=False,
    help="List workspaces created by me only.",
)
@click.option(
    "--state",
    "-s",
    "state_filter",
    multiple=True,
    callback=validate_workspace_state_filter,
    help=(
        "Filter by workspace state (repeatable). "
        f"Allowed: {', '.join(s.value for s in WorkspaceState)}"
    ),
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help="Include archived workspaces in the results.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help=(
        "Filter by tags. Can be repeated for multiple tags. "
        "Tags with the same key are ORed, different keys are ANDed. "
        "Example: --tag team:mlops --tag team:infra --tag env:prod filters to (team: mlops OR infra) AND env:prod."
    ),
)
@click.option(
    "--max-items",
    type=int,
    callback=validate_max_items,
    help="Max total items (only with --no-interactive).",
)
@click.option(
    "--page-size",
    type=int,
    default=10,
    show_default=True,
    callback=validate_page_size,
    help=f"Items per page (max {MAX_PAGE_SIZE}).",
)
@click.option(
    "--interactive/--no-interactive",
    default=True,
    show_default=True,
    help="Use interactive paging.",
)
@click.option(
    "--sort",
    help=(
        "Sort by FIELD (prefix with '-' for desc). "
        f"Allowed: {', '.join(f.value for f in WorkspaceSortField.__members__.values())}"
    ),
)
@click.option(
    "-j",
    "--json",
    "json_output",
    is_flag=True,
    default=False,
    help="Emit structured JSON to stdout.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Show all fields including IDs and metadata.",
)
def list(  # noqa: A001, PLR0912, PLR0913, PLR0917
    workspace_id: Optional[str],
    name: Optional[str],
    project: Optional[str],
    cloud: Optional[str],
    created_by_me: bool,
    state_filter: List[str],
    include_archived: bool,
    tags: Tuple[str],
    max_items: Optional[int],
    page_size: int,
    interactive: bool,
    sort: Optional[str],
    json_output: bool,
    verbose: bool,
) -> None:
    """List workspaces with optional filters."""
    if max_items is not None and interactive:
        raise click.UsageError("--max-items only allowed with --no-interactive")

    # Parse sort option
    sort_field, sort_order = _parse_sort_option(sort)

    stderr = Console(stderr=True)
    effective_max = max_items
    if not interactive and effective_max is None:
        stderr.print(
            f"Defaulting to {NON_INTERACTIVE_DEFAULT_MAX_ITEMS} items in batch mode; "
            "use --max-items to override."
        )
        effective_max = NON_INTERACTIVE_DEFAULT_MAX_ITEMS

    console = Console()
    creator_id = None
    if created_by_me:
        auth_block = get_auth_api_client()
        creator_id = auth_block.api_client.get_user_info_api_v2_userinfo_get().result.id

    # Parse tags filter
    tags_filter = parse_repeatable_tags_to_dict(tags) if tags else None

    # Convert state filter strings to WorkspaceState enums
    workspace_states = None
    if state_filter:
        workspace_states = [WorkspaceState.validate(s) for s in state_filter]

    stderr.print("[bold]Listing workspaces with:[/]")
    stderr.print(f"• name              = {name or '<any>'}")
    stderr.print(f"• project           = {project or '<any>'}")
    stderr.print(f"• cloud             = {cloud or '<any>'}")
    stderr.print(f"• created_by_me     = {created_by_me}")
    stderr.print(f"• states            = {', '.join(state_filter) or '<all>'}")
    stderr.print(f"• tags              = {tags_filter or '<none>'}")
    stderr.print(f"• include_archived  = {include_archived}")
    stderr.print(f"• sort              = {sort or '<none>'}")
    stderr.print(f"• mode                = {'interactive' if interactive else 'batch'}")
    stderr.print(f"• per-page limit      = {page_size}")
    stderr.print(f"• max-items total     = {effective_max or 'all'}")
    stderr.print(f"\nView your Workspaces in the UI at {get_endpoint('/workspaces')}\n")

    if json_output:

        def formatter(workspace):
            return workspace.to_dict()

        table_creator = _create_workspace_list_table
    elif verbose:
        formatter = _format_workspace_output_verbose
        table_creator = _create_workspace_list_table_verbose
    else:
        formatter = _format_workspace_output
        table_creator = _create_workspace_list_table

    try:
        iterator = anyscale.workspace.list(
            workspace_id=workspace_id,
            name=name,
            project=project,
            cloud=cloud,
            creator_id=creator_id,
            state_filter=workspace_states,
            tags_filter=tags_filter,
            include_config=False,  # CLI only needs metadata, not expensive config
            sort_field=sort_field,
            sort_order=sort_order,
            max_items=None if interactive else effective_max,
            page_size=page_size,
        )
        total = display_list(
            iterator=iter(iterator),
            item_formatter=formatter,
            table_creator=table_creator,
            json_output=json_output,
            page_size=page_size,
            interactive=interactive,
            max_items=effective_max,
            console=console,
        )

        if not json_output:
            if total > 0:
                stderr.print(f"\nFetched {total} workspaces.")
            else:
                stderr.print("\nNo workspaces found.")
    except ValueError as e:
        raise click.ClickException(str(e)) from None
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(f"Failed to list workspaces: {e}") from None



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ComputeTemplateQuery(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'orgwide': 'bool',
        'project_id': 'str',
        'creator_id': 'str',
        'name': 'TextQuery',
        'include_anonymous': 'bool',
        'archive_status': 'ArchiveStatus',
        'cloud_id': 'str',
        'version': 'int',
        'sort_by_clauses': 'list[SortByClause]'
    }

    attribute_map = {
        'orgwide': 'orgwide',
        'project_id': 'project_id',
        'creator_id': 'creator_id',
        'name': 'name',
        'include_anonymous': 'include_anonymous',
        'archive_status': 'archive_status',
        'cloud_id': 'cloud_id',
        'version': 'version',
        'sort_by_clauses': 'sort_by_clauses'
    }

    def __init__(self, orgwide=False, project_id=None, creator_id=None, name=None, include_anonymous=False, archive_status=None, cloud_id=None, version=None, sort_by_clauses=None, local_vars_configuration=None):  # noqa: E501
        """ComputeTemplateQuery - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._orgwide = None
        self._project_id = None
        self._creator_id = None
        self._name = None
        self._include_anonymous = None
        self._archive_status = None
        self._cloud_id = None
        self._version = None
        self._sort_by_clauses = None
        self.discriminator = None

        if orgwide is not None:
            self.orgwide = orgwide
        if project_id is not None:
            self.project_id = project_id
        if creator_id is not None:
            self.creator_id = creator_id
        if name is not None:
            self.name = name
        if include_anonymous is not None:
            self.include_anonymous = include_anonymous
        if archive_status is not None:
            self.archive_status = archive_status
        if cloud_id is not None:
            self.cloud_id = cloud_id
        if version is not None:
            self.version = version
        if sort_by_clauses is not None:
            self.sort_by_clauses = sort_by_clauses

    @property
    def orgwide(self):
        """Gets the orgwide of this ComputeTemplateQuery.  # noqa: E501

        This option is deprectated. It does nothing.  # noqa: E501

        :return: The orgwide of this ComputeTemplateQuery.  # noqa: E501
        :rtype: bool
        """
        return self._orgwide

    @orgwide.setter
    def orgwide(self, orgwide):
        """Sets the orgwide of this ComputeTemplateQuery.

        This option is deprectated. It does nothing.  # noqa: E501

        :param orgwide: The orgwide of this ComputeTemplateQuery.  # noqa: E501
        :type: bool
        """

        self._orgwide = orgwide

    @property
    def project_id(self):
        """Gets the project_id of this ComputeTemplateQuery.  # noqa: E501


        :return: The project_id of this ComputeTemplateQuery.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this ComputeTemplateQuery.


        :param project_id: The project_id of this ComputeTemplateQuery.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def creator_id(self):
        """Gets the creator_id of this ComputeTemplateQuery.  # noqa: E501

        Filters Compute Templates by creator. This is only supported when `orgwide` is True.  # noqa: E501

        :return: The creator_id of this ComputeTemplateQuery.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this ComputeTemplateQuery.

        Filters Compute Templates by creator. This is only supported when `orgwide` is True.  # noqa: E501

        :param creator_id: The creator_id of this ComputeTemplateQuery.  # noqa: E501
        :type: str
        """

        self._creator_id = creator_id

    @property
    def name(self):
        """Gets the name of this ComputeTemplateQuery.  # noqa: E501

        Filters Compute Templates by name. If this field is absent, no filtering is done.  # noqa: E501

        :return: The name of this ComputeTemplateQuery.  # noqa: E501
        :rtype: TextQuery
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ComputeTemplateQuery.

        Filters Compute Templates by name. If this field is absent, no filtering is done.  # noqa: E501

        :param name: The name of this ComputeTemplateQuery.  # noqa: E501
        :type: TextQuery
        """

        self._name = name

    @property
    def include_anonymous(self):
        """Gets the include_anonymous of this ComputeTemplateQuery.  # noqa: E501

        Whether to include anonymous Compute Templates in the search.Anonymous compute templates are usually not shown in list views.  # noqa: E501

        :return: The include_anonymous of this ComputeTemplateQuery.  # noqa: E501
        :rtype: bool
        """
        return self._include_anonymous

    @include_anonymous.setter
    def include_anonymous(self, include_anonymous):
        """Sets the include_anonymous of this ComputeTemplateQuery.

        Whether to include anonymous Compute Templates in the search.Anonymous compute templates are usually not shown in list views.  # noqa: E501

        :param include_anonymous: The include_anonymous of this ComputeTemplateQuery.  # noqa: E501
        :type: bool
        """

        self._include_anonymous = include_anonymous

    @property
    def archive_status(self):
        """Gets the archive_status of this ComputeTemplateQuery.  # noqa: E501

        The archive status to filter by. Defaults to unarchived.  # noqa: E501

        :return: The archive_status of this ComputeTemplateQuery.  # noqa: E501
        :rtype: ArchiveStatus
        """
        return self._archive_status

    @archive_status.setter
    def archive_status(self, archive_status):
        """Sets the archive_status of this ComputeTemplateQuery.

        The archive status to filter by. Defaults to unarchived.  # noqa: E501

        :param archive_status: The archive_status of this ComputeTemplateQuery.  # noqa: E501
        :type: ArchiveStatus
        """

        self._archive_status = archive_status

    @property
    def cloud_id(self):
        """Gets the cloud_id of this ComputeTemplateQuery.  # noqa: E501

        Filters Compute Templates by cloud  # noqa: E501

        :return: The cloud_id of this ComputeTemplateQuery.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this ComputeTemplateQuery.

        Filters Compute Templates by cloud  # noqa: E501

        :param cloud_id: The cloud_id of this ComputeTemplateQuery.  # noqa: E501
        :type: str
        """

        self._cloud_id = cloud_id

    @property
    def version(self):
        """Gets the version of this ComputeTemplateQuery.  # noqa: E501

        Filters Compute Templates by version. Versions are positive integers. Setting this field to -1 will return only the latest version of each Compute Template. Setting this field to -2 will not filter by version. Setting version to None is equivalent to setting version to '-1'.  # noqa: E501

        :return: The version of this ComputeTemplateQuery.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this ComputeTemplateQuery.

        Filters Compute Templates by version. Versions are positive integers. Setting this field to -1 will return only the latest version of each Compute Template. Setting this field to -2 will not filter by version. Setting version to None is equivalent to setting version to '-1'.  # noqa: E501

        :param version: The version of this ComputeTemplateQuery.  # noqa: E501
        :type: int
        """

        self._version = version

    @property
    def sort_by_clauses(self):
        """Gets the sort_by_clauses of this ComputeTemplateQuery.  # noqa: E501

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :return: The sort_by_clauses of this ComputeTemplateQuery.  # noqa: E501
        :rtype: list[SortByClause]
        """
        return self._sort_by_clauses

    @sort_by_clauses.setter
    def sort_by_clauses(self, sort_by_clauses):
        """Sets the sort_by_clauses of this ComputeTemplateQuery.

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :param sort_by_clauses: The sort_by_clauses of this ComputeTemplateQuery.  # noqa: E501
        :type: list[SortByClause]
        """

        self._sort_by_clauses = sort_by_clauses

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ComputeTemplateQuery):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ComputeTemplateQuery):
            return True

        return self.to_dict() != other.to_dict()



import inspect
import sys
from types import ModuleType
from typing import Any, Dict, List, Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.cloud._private.cloud_sdk import PrivateCloudSDK
from anyscale.cloud.commands import (
    _ADD_COLLABORATORS_ARG_DOCSTRINGS,
    _ADD_COLLABORATORS_EXAMPLE,
    _GET_ARG_DOCSTRINGS,
    _GET_DEFAULT_EXAMPLE,
    _GET_EXAMPLE,
    _TERMINATE_SYSTEM_CLUSTER_ARG_DOCSTRINGS,
    _TERMINATE_SYSTEM_CLUSTER_EXAMPLE,
    add_collaborators,
    get,
    get_default,
    list,  # noqa: A004 - claude_comment("claude-opus-4-5", "SDK public API re-export")
    terminate_system_cluster,
)
from anyscale.cloud.models import Cloud, CreateCloudCollaborator
from anyscale.connect import ClientBuilder


class CloudSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateCloudSDK(client=client, logger=logger, timer=timer)

    @sdk_docs(
        doc_py_example=_ADD_COLLABORATORS_EXAMPLE,
        arg_docstrings=_ADD_COLLABORATORS_ARG_DOCSTRINGS,
    )
    def add_collaborators(
        self, cloud: str, collaborators: List[CreateCloudCollaborator],
    ) -> None:
        """
        Batch add collaborators to a cloud.

        :param cloud: The cloud to add users to.
        :param collaborators: The list of collaborators to add to the cloud.
        """
        self._private_sdk.add_collaborators(cloud, collaborators)

    @sdk_docs(
        doc_py_example=_GET_EXAMPLE, arg_docstrings=_GET_ARG_DOCSTRINGS,
    )
    def get_cloud(
        self, id: Optional[str], name: Optional[str],  # noqa: A002
    ) -> Optional[Cloud]:
        """
        Retrieve a cloud by its name or ID.

        :param id: The ID of the cloud to retrieve.
        :param name: The name of the cloud to retrieve.
        :return: A ``Cloud`` object if found, otherwise ``None``.
        """
        return self._private_sdk.get(id=id, name=name)

    @sdk_docs(
        doc_py_example=_GET_DEFAULT_EXAMPLE, arg_docstrings={},
    )
    def get_default_cloud(self) -> Optional[Cloud]:
        """
        Get the default cloud for your organization.

        :return: The default ``Cloud`` object if it exists, otherwise ``None``.
        """
        return self._private_sdk.get_default()

    @sdk_docs(
        doc_py_example=_TERMINATE_SYSTEM_CLUSTER_EXAMPLE,
        arg_docstrings=_TERMINATE_SYSTEM_CLUSTER_ARG_DOCSTRINGS,
    )
    def terminate_system_cluster(self, cloud_id: str, wait: bool = False) -> str:
        """
        Terminate the system cluster for a given cloud.

        :return: ID of the terminated system cluster.
        """
        return self._private_sdk.terminate_system_cluster(cloud_id, wait)


# Note: indentation here matches that of connect.py::ClientBuilder.
BUILDER_HELP_FOOTER = """
        See ``anyscale.ClientBuilder`` for full documentation of
        this experimental feature."""


class CloudModule(ModuleType):
    """
    A custom callable module object for `anyscale.cloud`.

    This hack is needed since `anyscale.cloud` is a function for Anyscale connect
    but also a module for the SDK.
    """

    def __init__(self):
        # Expose attributes from the SDK.
        self.CloudSDK = CloudSDK
        self.add_collaborators = add_collaborators
        self.list = list
        self.get = get
        self.get_default = get_default
        self.terminate_system_cluster = terminate_system_cluster

        # Expose Anyscale connect
        self.new_builder = self._new_builder()

    # This code is copied from frontend/cli/anyscale/__init__.py.
    def _new_builder(self) -> Any:
        target = ClientBuilder.cloud

        def new_session_builder(*a: List[Any], **kw: Dict[str, Any]) -> Any:
            builder = ClientBuilder()
            return target(builder, *a, **kw)  # type: ignore

        new_session_builder.__name__ = "cloud"
        new_session_builder.__doc__ = target.__doc__ + BUILDER_HELP_FOOTER  # type: ignore
        new_session_builder.__signature__ = inspect.signature(target)  # type: ignore

        return new_session_builder

    def __call__(self, *args, **kwargs):
        """
        Define the behavior when `anyscale.cloud` is called for Anyscale connect.
        """
        return self.new_builder(*args, **kwargs)


sys.modules[__name__] = CloudModule()



from typing import List

from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.user_group.models import UserGroup


class PrivateUserGroupSDK(BaseSDK):
    """Private SDK for user group operations."""

    def list(self, max_items: int = 50) -> List[UserGroup]:
        """
        List user groups in the organization.

        Args:
            max_items: Maximum number of user groups to return.

        Returns:
            List of UserGroup objects.
        """
        response = self.client.list_user_groups(count=max_items)
        if response.results is None:
            return []
        return [UserGroup.from_api_model(ug) for ug in response.results]

    def get(self, group_id: str) -> UserGroup:
        """
        Get a specific user group by ID.

        Args:
            group_id: The ID of the user group.

        Returns:
            UserGroup object.
        """
        api_model = self.client.get_user_group(group_id)
        return UserGroup.from_api_model(api_model)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class TrainRunAttempt(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'run_id': 'str',
        'attempt_id': 'str',
        'status': 'RunAttemptStatus',
        'status_detail': 'str',
        'start_time_ms': 'int',
        'end_time_ms': 'int',
        'resources': 'list[TrainResources]',
        'workers': 'list[TrainWorker]'
    }

    attribute_map = {
        'run_id': 'run_id',
        'attempt_id': 'attempt_id',
        'status': 'status',
        'status_detail': 'status_detail',
        'start_time_ms': 'start_time_ms',
        'end_time_ms': 'end_time_ms',
        'resources': 'resources',
        'workers': 'workers'
    }

    def __init__(self, run_id=None, attempt_id=None, status=None, status_detail=None, start_time_ms=None, end_time_ms=None, resources=None, workers=None, local_vars_configuration=None):  # noqa: E501
        """TrainRunAttempt - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._run_id = None
        self._attempt_id = None
        self._status = None
        self._status_detail = None
        self._start_time_ms = None
        self._end_time_ms = None
        self._resources = None
        self._workers = None
        self.discriminator = None

        self.run_id = run_id
        self.attempt_id = attempt_id
        self.status = status
        if status_detail is not None:
            self.status_detail = status_detail
        self.start_time_ms = start_time_ms
        if end_time_ms is not None:
            self.end_time_ms = end_time_ms
        self.resources = resources
        self.workers = workers

    @property
    def run_id(self):
        """Gets the run_id of this TrainRunAttempt.  # noqa: E501


        :return: The run_id of this TrainRunAttempt.  # noqa: E501
        :rtype: str
        """
        return self._run_id

    @run_id.setter
    def run_id(self, run_id):
        """Sets the run_id of this TrainRunAttempt.


        :param run_id: The run_id of this TrainRunAttempt.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and run_id is None:  # noqa: E501
            raise ValueError("Invalid value for `run_id`, must not be `None`")  # noqa: E501

        self._run_id = run_id

    @property
    def attempt_id(self):
        """Gets the attempt_id of this TrainRunAttempt.  # noqa: E501


        :return: The attempt_id of this TrainRunAttempt.  # noqa: E501
        :rtype: str
        """
        return self._attempt_id

    @attempt_id.setter
    def attempt_id(self, attempt_id):
        """Sets the attempt_id of this TrainRunAttempt.


        :param attempt_id: The attempt_id of this TrainRunAttempt.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and attempt_id is None:  # noqa: E501
            raise ValueError("Invalid value for `attempt_id`, must not be `None`")  # noqa: E501

        self._attempt_id = attempt_id

    @property
    def status(self):
        """Gets the status of this TrainRunAttempt.  # noqa: E501


        :return: The status of this TrainRunAttempt.  # noqa: E501
        :rtype: RunAttemptStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this TrainRunAttempt.


        :param status: The status of this TrainRunAttempt.  # noqa: E501
        :type: RunAttemptStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def status_detail(self):
        """Gets the status_detail of this TrainRunAttempt.  # noqa: E501


        :return: The status_detail of this TrainRunAttempt.  # noqa: E501
        :rtype: str
        """
        return self._status_detail

    @status_detail.setter
    def status_detail(self, status_detail):
        """Sets the status_detail of this TrainRunAttempt.


        :param status_detail: The status_detail of this TrainRunAttempt.  # noqa: E501
        :type: str
        """

        self._status_detail = status_detail

    @property
    def start_time_ms(self):
        """Gets the start_time_ms of this TrainRunAttempt.  # noqa: E501


        :return: The start_time_ms of this TrainRunAttempt.  # noqa: E501
        :rtype: int
        """
        return self._start_time_ms

    @start_time_ms.setter
    def start_time_ms(self, start_time_ms):
        """Sets the start_time_ms of this TrainRunAttempt.


        :param start_time_ms: The start_time_ms of this TrainRunAttempt.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and start_time_ms is None:  # noqa: E501
            raise ValueError("Invalid value for `start_time_ms`, must not be `None`")  # noqa: E501

        self._start_time_ms = start_time_ms

    @property
    def end_time_ms(self):
        """Gets the end_time_ms of this TrainRunAttempt.  # noqa: E501


        :return: The end_time_ms of this TrainRunAttempt.  # noqa: E501
        :rtype: int
        """
        return self._end_time_ms

    @end_time_ms.setter
    def end_time_ms(self, end_time_ms):
        """Sets the end_time_ms of this TrainRunAttempt.


        :param end_time_ms: The end_time_ms of this TrainRunAttempt.  # noqa: E501
        :type: int
        """

        self._end_time_ms = end_time_ms

    @property
    def resources(self):
        """Gets the resources of this TrainRunAttempt.  # noqa: E501


        :return: The resources of this TrainRunAttempt.  # noqa: E501
        :rtype: list[TrainResources]
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """Sets the resources of this TrainRunAttempt.


        :param resources: The resources of this TrainRunAttempt.  # noqa: E501
        :type: list[TrainResources]
        """
        if self.local_vars_configuration.client_side_validation and resources is None:  # noqa: E501
            raise ValueError("Invalid value for `resources`, must not be `None`")  # noqa: E501

        self._resources = resources

    @property
    def workers(self):
        """Gets the workers of this TrainRunAttempt.  # noqa: E501


        :return: The workers of this TrainRunAttempt.  # noqa: E501
        :rtype: list[TrainWorker]
        """
        return self._workers

    @workers.setter
    def workers(self, workers):
        """Sets the workers of this TrainRunAttempt.


        :param workers: The workers of this TrainRunAttempt.  # noqa: E501
        :type: list[TrainWorker]
        """
        if self.local_vars_configuration.client_side_validation and workers is None:  # noqa: E501
            raise ValueError("Invalid value for `workers`, must not be `None`")  # noqa: E501

        self._workers = workers

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TrainRunAttempt):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, TrainRunAttempt):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class UsageByCloud(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'anyscale_credits': 'float',
        'dollar_value': 'float',
        'date': 'date',
        'cloud_id': 'str',
        'cloud_name': 'str'
    }

    attribute_map = {
        'anyscale_credits': 'anyscale_credits',
        'dollar_value': 'dollar_value',
        'date': 'date',
        'cloud_id': 'cloud_id',
        'cloud_name': 'cloud_name'
    }

    def __init__(self, anyscale_credits=None, dollar_value=None, date=None, cloud_id=None, cloud_name=None, local_vars_configuration=None):  # noqa: E501
        """UsageByCloud - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._anyscale_credits = None
        self._dollar_value = None
        self._date = None
        self._cloud_id = None
        self._cloud_name = None
        self.discriminator = None

        self.anyscale_credits = anyscale_credits
        if dollar_value is not None:
            self.dollar_value = dollar_value
        if date is not None:
            self.date = date
        self.cloud_id = cloud_id
        self.cloud_name = cloud_name

    @property
    def anyscale_credits(self):
        """Gets the anyscale_credits of this UsageByCloud.  # noqa: E501


        :return: The anyscale_credits of this UsageByCloud.  # noqa: E501
        :rtype: float
        """
        return self._anyscale_credits

    @anyscale_credits.setter
    def anyscale_credits(self, anyscale_credits):
        """Sets the anyscale_credits of this UsageByCloud.


        :param anyscale_credits: The anyscale_credits of this UsageByCloud.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and anyscale_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `anyscale_credits`, must not be `None`")  # noqa: E501

        self._anyscale_credits = anyscale_credits

    @property
    def dollar_value(self):
        """Gets the dollar_value of this UsageByCloud.  # noqa: E501


        :return: The dollar_value of this UsageByCloud.  # noqa: E501
        :rtype: float
        """
        return self._dollar_value

    @dollar_value.setter
    def dollar_value(self, dollar_value):
        """Sets the dollar_value of this UsageByCloud.


        :param dollar_value: The dollar_value of this UsageByCloud.  # noqa: E501
        :type: float
        """

        self._dollar_value = dollar_value

    @property
    def date(self):
        """Gets the date of this UsageByCloud.  # noqa: E501


        :return: The date of this UsageByCloud.  # noqa: E501
        :rtype: date
        """
        return self._date

    @date.setter
    def date(self, date):
        """Sets the date of this UsageByCloud.


        :param date: The date of this UsageByCloud.  # noqa: E501
        :type: date
        """

        self._date = date

    @property
    def cloud_id(self):
        """Gets the cloud_id of this UsageByCloud.  # noqa: E501


        :return: The cloud_id of this UsageByCloud.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this UsageByCloud.


        :param cloud_id: The cloud_id of this UsageByCloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def cloud_name(self):
        """Gets the cloud_name of this UsageByCloud.  # noqa: E501


        :return: The cloud_name of this UsageByCloud.  # noqa: E501
        :rtype: str
        """
        return self._cloud_name

    @cloud_name.setter
    def cloud_name(self, cloud_name):
        """Sets the cloud_name of this UsageByCloud.


        :param cloud_name: The cloud_name of this UsageByCloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_name is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_name`, must not be `None`")  # noqa: E501

        self._cloud_name = cloud_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UsageByCloud):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UsageByCloud):
            return True

        return self.to_dict() != other.to_dict()



from __future__ import annotations

from datetime import datetime
from enum import Enum
from functools import partial
from json import dumps as json_dumps
import sys
from typing import Dict, get_type_hints, List, Optional, Tuple

import click
from rich.console import Console
from rich.table import Table

from anyscale.client.openapi_client.models.job_queue_sort_directive import (
    JobQueueSortDirective,
)
from anyscale.client.openapi_client.models.job_queue_sort_field import JobQueueSortField
from anyscale.client.openapi_client.models.session_state import SessionState
from anyscale.client.openapi_client.models.sort_order import SortOrder
from anyscale.commands import command_examples
from anyscale.commands.list_util import (
    display_list,
    MAX_PAGE_SIZE,
    NON_INTERACTIVE_DEFAULT_MAX_ITEMS,
    validate_page_size,
)
from anyscale.commands.util import (
    AnyscaleCommand,
    build_kv_table,
    parse_repeatable_tags_to_dict,
    parse_tags_kv_to_str_map,
)
import anyscale.job_queue
from anyscale.job_queue.models import JobQueueStatus, JobQueueStatusKeys
from anyscale.util import get_endpoint, get_user_info, validate_non_negative_arg


@click.group("job-queue", help="Manage Anyscale Job Queues.")
def job_queue_cli() -> None:
    pass


class ViewOption(Enum):
    DEFAULT = "default"
    STATS = "stats"
    ALL = "all"


VIEW_COLUMNS: Dict[ViewOption, List[JobQueueStatusKeys]] = {
    ViewOption.DEFAULT: [
        "name",
        "id",
        "state",
        "creator_email",
        "project_id",
        "created_at",
    ],
    ViewOption.STATS: [
        "id",
        "name",
        "total_jobs",
        "active_jobs",
        "successful_jobs",
        "failed_jobs",
    ],
    ViewOption.ALL: [
        "name",
        "id",
        "state",
        "creator_email",
        "project_id",
        "created_at",
        "max_concurrency",
        "idle_timeout_s",
        "cloud_id",
        "user_provided_id",
        "execution_mode",
        "total_jobs",
        "active_jobs",
        "successful_jobs",
        "failed_jobs",
    ],
}


@job_queue_cli.command(
    name="list",
    help="List job queues.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_LIST,
)
@click.option("--id", "job_queue_id", help="ID of a job queue.")
@click.option("--name", type=str, help="Filter by name.")
@click.option("--cloud", type=str, help="Filter by cloud.")
@click.option("--project", type=str, help="Filter by project.")
@click.option(
    "--include-all-users/--only-mine",
    default=False,
    help="Include job queues not created by current user.",
)
@click.option(
    "--cluster-status",
    type=click.Choice(SessionState.allowable_values, case_sensitive=False),
    help="Filter by cluster status.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help=(
        "This option can be repeated to filter by multiple tags. "
        "Tags with the same key are ORed, whereas tags with different keys are ANDed. "
        "Example: --tag team:mlops --tag team:infra --tag env:prod. "
        "Filters with team: (mlops OR infra) AND env:prod."
    ),
)
@click.option(
    "--view",
    type=click.Choice([opt.value for opt in ViewOption], case_sensitive=False),
    default=ViewOption.DEFAULT.value,
    help="Columns view.",
    callback=lambda _ctx, _param, value: ViewOption(value),
)
@click.option(
    "--page-size",
    default=10,
    type=int,
    callback=validate_page_size,
    help=f"Items per page (max {MAX_PAGE_SIZE}).",
)
@click.option(
    "--max-items",
    type=int,
    callback=lambda ctx, param, value: validate_non_negative_arg(ctx, param, value)
    if value
    else None,
    help="Non-interactive max items.",
)
@click.option(
    "--sort",
    "sort_dirs",
    multiple=True,
    default=["-created_at"],
    help="Sort by FIELD (prefix with '-' for desc). Repeatable.",
    callback=lambda _ctx, _param, values: _parse_sort_fields("sort", list(values)),
)
@click.option(
    "--no-interactive/--interactive",
    default=False,
    help="Use non-interactive batch mode instead of interactive paging.",
)
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="JSON output.",
)
def list_job_queues(  # noqa: PLR0913
    job_queue_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    cluster_status: Optional[str],
    tags: List[str],
    include_all_users: bool,
    view: ViewOption,
    page_size: int,
    max_items: Optional[int],
    sort_dirs: List[JobQueueSortDirective],
    no_interactive: bool,
    json_output: bool,
) -> None:
    """List and page job queues according to filters and view."""
    if max_items and not no_interactive:
        raise click.UsageError("--max-items only in non-interactive mode")

    effective_max = max_items or NON_INTERACTIVE_DEFAULT_MAX_ITEMS
    console = Console()
    stderr = Console(stderr=True)

    if not json_output:
        _print_list_diagnostics(
            stderr=stderr,
            job_queue_id=job_queue_id,
            name=name,
            include_all_users=include_all_users,
            cloud=cloud,
            project=project,
            cluster_status=cluster_status,
            view=view,
            sort_dirs=sort_dirs,
            no_interactive=no_interactive,
            page_size=page_size,
            effective_max=effective_max,
        )

    try:
        user = get_user_info()
        iterator = anyscale.job_queue.list(
            job_queue_id=job_queue_id,
            name=name,
            creator_id=None if include_all_users else (user.id if user else None),
            cloud=cloud,
            project=project,
            tags_filter=parse_repeatable_tags_to_dict(tags) if tags else None,
            page_size=page_size,
            max_items=None if not no_interactive else effective_max,
            sorting_directives=sort_dirs,
        )
        cols = VIEW_COLUMNS[view]
        table_fn = partial(_create_table, view)

        def row_fn(jq: JobQueueStatus) -> Dict[str, str]:
            data = _format_data(jq)
            return {c: data[c] for c in cols}

        total = display_list(
            iterator=iter(iterator),
            item_formatter=row_fn,
            table_creator=table_fn,
            json_output=json_output,
            page_size=page_size,
            interactive=not no_interactive,
            max_items=effective_max,
            console=console,
        )
        if not json_output:
            stderr.print(f"Fetched {total} queues" if total else "No queues found.")

    except Exception as e:  # noqa: BLE001
        stderr.print(f"Error: {e}", style="red")
        sys.exit(1)


@job_queue_cli.command(
    name="update",
    help="Update job queue settings.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_UPDATE,
)
@click.option("--id", "job_queue_id", required=True, help="ID of the job queue.")
@click.option("--max-concurrency", type=int, help="Max number of concurrent jobs.")
@click.option("--idle-timeout-s", type=int, help="Idle timeout in seconds.")
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="JSON output.",
)
def update_job_queue(
    job_queue_id: str,
    max_concurrency: Optional[int],
    idle_timeout_s: Optional[int],
    json_output: bool,
) -> None:
    """Update the max_concurrency or idle_timeout_s of a job queue."""
    if max_concurrency is None and idle_timeout_s is None:
        raise click.ClickException("Specify --max-concurrency or --idle-timeout-s")
    stderr = Console(stderr=True)
    if not json_output:
        stderr.print(f"Updating job queue '{job_queue_id}'...")
    try:
        jq = anyscale.job_queue.update(
            job_queue_id=job_queue_id,
            job_queue_name=None,
            max_concurrency=max_concurrency,
            idle_timeout_s=idle_timeout_s,
        )
        if json_output:
            Console().print_json(json_dumps(_format_data(jq), indent=2))
        else:
            _display_single(jq, stderr, ViewOption.ALL)
    except Exception as e:  # noqa: BLE001
        stderr.print(f"Update failed: {e}", style="red")
        sys.exit(1)


@job_queue_cli.group("tags", help="Manage tags for job queues.")
def job_queue_tags_cli() -> None:
    pass


@job_queue_tags_cli.command(
    name="add",
    help="Add or update tags on a job queue.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_TAGS_ADD_EXAMPLE,
)
@click.option("--id", "job_queue_id", help="ID of a job queue.")
@click.option("--name", "-n", type=str, help="Name of a job queue.")
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
def add_tags(
    job_queue_id: Optional[str], name: Optional[str], tags: Tuple[str],
) -> None:
    if not job_queue_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = parse_tags_kv_to_str_map(tags)
    if not tag_map:
        raise click.ClickException("Provide at least one --tag key=value.")
    anyscale.job_queue.add_tags(job_queue_id=job_queue_id, name=name, tags=tag_map)
    stderr = Console(stderr=True)
    ident = job_queue_id or name or "<unknown>"
    stderr.print(f"Tags updated for job queue '{ident}'.")


@job_queue_tags_cli.command(
    name="remove",
    help="Remove tags by key from a job queue.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_TAGS_REMOVE_EXAMPLE,
)
@click.option("--id", "job_queue_id", help="ID of a job queue.")
@click.option("--name", "-n", type=str, help="Name of a job queue.")
@click.option("--key", "keys", multiple=True, help="Tag key to remove. Repeatable.")
def remove_tags(
    job_queue_id: Optional[str], name: Optional[str], keys: Tuple[str],
) -> None:
    if not job_queue_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    key_list = [k for k in keys if k and k.strip()]
    if not key_list:
        raise click.ClickException("Provide at least one --key to remove.")
    anyscale.job_queue.remove_tags(job_queue_id=job_queue_id, name=name, keys=key_list)
    stderr = Console(stderr=True)
    ident = job_queue_id or name or "<unknown>"
    stderr.print(f"Removed tag keys {key_list} from job queue '{ident}'.")


@job_queue_tags_cli.command(
    name="list",
    help="List tags for a job queue.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_TAGS_LIST_EXAMPLE,
)
@click.option("--id", "job_queue_id", help="ID of a job queue.")
@click.option("--name", "-n", type=str, help="Name of a job queue.")
@click.option("--json", "json_output", is_flag=True, default=False, help="JSON output.")
def list_tags(
    job_queue_id: Optional[str], name: Optional[str], json_output: bool,
) -> None:
    if not job_queue_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = anyscale.job_queue.list_tags(job_queue_id=job_queue_id, name=name)
    if json_output:
        Console().print_json(json=json_dumps(tag_map, indent=2))
    else:
        stderr = Console(stderr=True)
        if not tag_map:
            stderr.print("No tags found.")
            return
        pairs = tag_map.items()
        stderr.print(build_kv_table(pairs, title="Tags"))


@job_queue_cli.command(
    name="status",
    help="Show job queue details.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_QUEUE_STATUS,
)
@click.option("--id", "job_queue_id", required=True, help="ID of the job queue.")
@click.option(
    "--view",
    type=click.Choice([opt.value for opt in ViewOption], case_sensitive=False),
    default=ViewOption.DEFAULT.value,
    help="Columns view.",
    callback=lambda _ctx, _param, value: ViewOption(value),
)
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="JSON output.",
)
def status(job_queue_id: str, view: ViewOption, json_output: bool,) -> None:
    """Fetch and display a single job queue's details."""
    stderr = Console(stderr=True)
    if not json_output:
        stderr.print(f"Fetching job queue '{job_queue_id}'...")
    try:
        jq = anyscale.job_queue.status(job_queue_id=job_queue_id)
        if json_output:
            Console().print_json(json_dumps(_format_data(jq), indent=2))
        else:
            # Use ALL view for single item display if not specified
            display_view = ViewOption.ALL if view == ViewOption.DEFAULT else view
            _display_single(jq, stderr, display_view)
    except Exception as e:  # noqa: BLE001
        stderr.print(f"Failed: {e}", style="red")
        sys.exit(1)


def _parse_sort_fields(
    param: str, sort_fields: List[str],
) -> List[JobQueueSortDirective]:
    """Convert a list of string fields into JobQueueSortDirective objects."""
    directives: List[JobQueueSortDirective] = []
    opts = ", ".join(v.lower() for v in JobQueueSortField.allowable_values)
    for field_str in sort_fields:
        desc = field_str.startswith("-")
        raw = field_str.lstrip("-").upper()
        if raw not in JobQueueSortField.allowable_values:
            raise click.UsageError(f"{param} must be one of {opts}")
        directives.append(
            JobQueueSortDirective(
                sort_field=raw, sort_order=SortOrder.DESC if desc else SortOrder.ASC,
            )
        )
    return directives


def _create_table(view: ViewOption, show_header: bool) -> Table:
    """Create a Rich Table with columns based on the selected view."""
    table = Table(show_header=show_header, expand=True)
    for key in VIEW_COLUMNS[view]:
        table.add_column(key.replace("_", " ").upper(), overflow="fold")
    return table


def _format_data(jq: JobQueueStatus) -> Dict[str, str]:
    """Format a JobQueueStatus object into a dictionary of strings for display."""
    data = {}
    for key in get_type_hints(JobQueueStatus):
        value = getattr(jq, key, None)
        if isinstance(value, datetime):
            data[key] = value.strftime("%Y-%m-%d %H:%M:%S") if value else ""
        elif value is None:
            data[key] = ""
        else:
            data[key] = str(value)
    return data


def _display_single(jq: JobQueueStatus, stderr: Console, view: ViewOption,) -> None:
    """Display a single job queue's details in a table using the selected view.

    Args:
        jq: The JobQueueStatus object to display.
        stderr: The Rich Console object to print to.
        view: The ViewOption determining which columns to display.
    """
    table = _create_table(view, show_header=True)
    data = _format_data(jq)
    table.add_row(*(data[col] for col in VIEW_COLUMNS[view]))
    stderr.print(table)


def _print_list_diagnostics(  # noqa: PLR0913
    stderr: Console,
    job_queue_id: Optional[str],
    name: Optional[str],
    include_all_users: bool,
    cloud: Optional[str],
    project: Optional[str],
    cluster_status: Optional[str],
    view: ViewOption,
    sort_dirs: List[JobQueueSortDirective],
    no_interactive: bool,
    page_size: int,
    effective_max: int,
) -> None:
    """Prints diagnostic information for the list_job_queues command."""
    stderr.print("[bold]Listing with:[/]")
    stderr.print(f"id: {job_queue_id or '<any>'}")
    stderr.print(f"name: {name or '<any>'}")
    stderr.print(f"creator: {'all' if include_all_users else 'mine'}")
    stderr.print(f"cloud: {cloud or '<any>'}")
    stderr.print(f"project: {project or '<any>'}")
    stderr.print(f"cluster: {cluster_status or '<any>'}")
    stderr.print(f"view: {view.value}")

    formatted_sort_dirs = [
        f"{'-' if d.sort_order == SortOrder.DESC else ''}{(d.sort_field or '').lower()}"
        for d in sort_dirs
    ]
    stderr.print(f"sort: {formatted_sort_dirs}")

    stderr.print(f"mode: {'batch' if no_interactive else 'interactive'}")
    stderr.print(f"page-size: {page_size}")
    stderr.print(f"max-items: {effective_max}")
    stderr.print(f"UI: {get_endpoint('/job-queues')}\n")



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Union

from anyscale._private.models import ModelBase, ModelEnum


class CloudPermissionLevel(ModelEnum):
    """Permission levels for cloud collaborators."""

    WRITE = "WRITE"
    READONLY = "READONLY"

    __docstrings__: ClassVar[Dict[str, str]] = {
        WRITE: "Write permission level for the cloud",
        READONLY: "Readonly permission level for the cloud",
    }


@dataclass(frozen=True)
class CreateCloudCollaborator(ModelBase):
    """User to be added as a collaborator to a cloud.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.cloud.models import CloudPermissionLevel, CreateCloudCollaborator

create_cloud_collaborator = CreateCloudCollaborator(
   # Email of the user to be added as a collaborator
    email="test@anyscale.com",
    # Permission level for the user to the cloud (CloudPermissionLevel.WRITE, CloudPermissionLevel.READONLY)
    permission_level=CloudPermissionLevel.READONLY,
)
"""

    def _validate_email(self, email: str):
        if not isinstance(email, str):
            raise TypeError("Email must be a string.")

    email: str = field(
        metadata={"docstring": "Email of the user to be added as a collaborator."},
    )

    def _validate_permission_level(
        self, permission_level: CloudPermissionLevel
    ) -> CloudPermissionLevel:
        if isinstance(permission_level, str):
            return CloudPermissionLevel.validate(permission_level)  # type: ignore
        elif isinstance(permission_level, CloudPermissionLevel):
            return permission_level
        else:
            raise TypeError(
                f"'permission_level' must be a 'CloudPermissionLevel' (it is {type(permission_level)})."
            )

    permission_level: CloudPermissionLevel = field(  # type: ignore
        default=CloudPermissionLevel.READONLY,  # type: ignore
        metadata={
            "docstring": "Permission level the added user should have for the cloud"  # type: ignore
            f"(one of: {','.join([str(m.value) for m in CloudPermissionLevel])}",  # type: ignore
        },
    )


@dataclass(frozen=True)
class CreateCloudCollaborators(ModelBase):
    """List of users to be added as collaborators to a cloud.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.cloud.models import CloudPermissionLevel, CreateCloudCollaborator, CreateCloudCollaborators

create_cloud_collaborator = CreateCloudCollaborator(
   # Email of the user to be added as a collaborator
    email="test@anyscale.com",
    # Permission level for the user to the cloud (CloudPermissionLevel.WRITE, CloudPermissionLevel.READONLY)
    permission_level=CloudPermissionLevel.READONLY,
)
create_cloud_collaborators = CreateCloudCollaborators(
    collaborators=[create_cloud_collaborator]
)
"""

    collaborators: List[Dict[str, Any]] = field(
        metadata={
            "docstring": "List of users to be added as collaborators to a cloud."
        },
    )

    def _validate_collaborators(self, collaborators: List[Dict[str, Any]]):
        if not isinstance(collaborators, list):
            raise TypeError("Collaborators must be a list.")


class ComputeStack(ModelEnum):
    """Type of compute stack for the cloud."""

    UNKNOWN = "UNKNOWN"
    VM = "VM"
    K8S = "K8S"

    __docstrings__: ClassVar[Dict[str, str]] = {
        UNKNOWN: "Unknown compute stack.",
        VM: "Virtual machine-based compute stack.",
        K8S: "Kubernetes-based compute stack.",
    }


class CloudProvider(ModelEnum):
    """Cloud infrastructure provider."""

    UNKNOWN = "UNKNOWN"
    AWS = "AWS"
    GCP = "GCP"
    AZURE = "AZURE"
    GENERIC = "GENERIC"

    __docstrings__: ClassVar[Dict[str, str]] = {
        UNKNOWN: "Unknown cloud provider.",
        AWS: "Amazon Web Services.",
        GCP: "Google Cloud Platform.",
        AZURE: "Microsoft Azure.",
        GENERIC: "Generic cloud provider.",
    }


@dataclass(frozen=True)
class Cloud(ModelBase):
    """Minimal Cloud resource model."""

    __doc_py_example__ = """\
from datetime import datetime
from anyscale.cloud.models import Cloud, CloudProvider, ComputeStack

cloud = Cloud(
    name="my-cloud",
    id="cloud-123",
    provider="AWS",  # This will be validated as CloudProvider.AWS
    region="us-west-2",
    created_at=datetime.now(),
    is_default=True,
    compute_stack="VM"  # This will be validated as ComputeStack.VM
)
"""

    name: str = field(metadata={"docstring": "Name of this Cloud."})
    id: str = field(metadata={"docstring": "Unique identifier for this Cloud."})
    provider: Union[CloudProvider, str] = field(
        metadata={
            "docstring": "Cloud provider (AWS, GCP, AZURE, GENERIC) or UNKNOWN if not recognized."
        },
    )
    compute_stack: Union[ComputeStack, str] = field(
        metadata={
            "docstring": "The compute stack associated with this cloud's primary cloud resource, or UNKNOWN if not recognized."
        },
    )
    region: Optional[str] = field(
        default=None, metadata={"docstring": "Region for this Cloud."}
    )
    created_at: Optional[datetime] = field(
        default=None, metadata={"docstring": "When the Cloud was created."}
    )
    is_default: Optional[bool] = field(
        default=None, metadata={"docstring": "Whether this is the default cloud."}
    )
    is_aggregated_logs_enabled: Optional[bool] = field(
        default=None,
        metadata={"docstring": "Whether aggregated logs are enabled for this cloud."},
    )

    def _validate_name(self, name: str) -> str:
        if not isinstance(name, str) or not name.strip():
            raise ValueError("name must be a non-empty string")
        return name

    def _validate_id(self, id: str) -> str:  # noqa: A002
        if not isinstance(id, str) or not id.strip():
            raise ValueError("id must be a non-empty string")
        return id

    def _validate_provider(self, provider: Union[CloudProvider, str]) -> CloudProvider:
        if isinstance(provider, str):
            # This will raise a ValueError if the provider is unrecognized.
            provider = CloudProvider(provider)
        elif not isinstance(provider, CloudProvider):
            raise TypeError("'provider' must be a CloudProvider.")

        return provider

    def _validate_region(self, region: Optional[str]) -> Optional[str]:
        if region is not None and not isinstance(region, str):
            raise TypeError("region must be a string")
        return region

    def _validate_created_at(
        self, created_at: Optional[datetime]
    ) -> Optional[datetime]:
        if created_at is None:
            return None
        if not isinstance(created_at, datetime):
            raise TypeError("created_at must be a datetime object")
        return created_at

    def _validate_is_default(self, is_default: Optional[bool]) -> Optional[bool]:
        if is_default is not None and not isinstance(is_default, bool):
            raise TypeError("is_default must be a bool")
        return is_default

    def _validate_compute_stack(
        self, compute_stack: Union[ComputeStack, str]
    ) -> ComputeStack:
        if isinstance(compute_stack, str):
            # This will raise a ValueError if the compute_stack is unrecognized.
            compute_stack = ComputeStack(compute_stack)
        elif not isinstance(compute_stack, ComputeStack):
            raise TypeError("'compute_stack' must be a ComputeStack.")

        return compute_stack

    def _validate_is_aggregated_logs_enabled(
        self, is_aggregated_logs_enabled: Optional[bool]
    ) -> Optional[bool]:
        if is_aggregated_logs_enabled is not None and not isinstance(
            is_aggregated_logs_enabled, bool
        ):
            raise TypeError("is_aggregated_logs_enabled must be a bool")
        return is_aggregated_logs_enabled


class NetworkingMode(ModelEnum):
    """Networking mode for cloud resources."""

    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"

    __docstrings__: ClassVar[Dict[str, str]] = {
        PUBLIC: "Direct networking.",
        PRIVATE: "Customer-defined networking.",
    }


@dataclass(frozen=True)
class NFSMountTarget(ModelBase):
    """NFS mount target configuration."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
nfs_mount_target:
  address: 123.456.789.012
"""

    address: str = field(metadata={"docstring": "The address of the NFS mount target."})
    zone: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The zone of the NFS mount target. If not set, this mount target may be used in any zone."
        },
    )


@dataclass(frozen=True)
class ObjectStorage(ModelBase):
    """Object storage configuration."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
object_storage:
  bucket_name: s3://my-bucket
"""

    bucket_name: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The cloud storage bucket name, prefixed with the storage scheme (s3://bucket-name, gs://bucket-name, or abfss://bucket-name@account.dfs.core.windows.net)."
        },
    )
    region: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The region for the cloud storage bucket. Defaults to the region of the cloud resource."
        },
    )
    endpoint: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The cloud storage endpoint, used to override the default cloud storage scheme's endpoint. For example, for S3, this will be passed to the AWS_ENDPOINT_URL environment variable."
        },
    )


@dataclass(frozen=True)
class FileStorage(ModelBase):
    """File storage configuration."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
file_storage:
  file_storage_id: fs-12345678901234567
"""

    file_storage_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "For AWS, the EFS ID. For GCP, the Filestore instance name."
        },
    )
    mount_targets: Optional[List[NFSMountTarget]] = field(
        default=None, metadata={"docstring": "The mount target(s) to use."}
    )
    mount_path: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "For GCP, the Filestore root directory. For NFS, the path of the server to mount from (e.g., <mount-target-address>/<mount-path> will be mounted)."
        },
    )
    persistent_volume_claim: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "For Kubernetes resources, the name of the persistent volume claim used to mount shared storage into pods."
        },
    )
    csi_ephemeral_volume_driver: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "For Kubernetes resources, the CSI ephemeral volume driver used to mount shared storage into pods."
        },
    )


@dataclass(frozen=True)
class AWSConfig(ModelBase):
    """AWS provider-specific configurations."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
aws_config:
  vpc_id: vpc-12345678901234567
  subnet_ids:
    - subnet-11111111111111111
    - subnet-22222222222222222
  security_group_ids:
    - sg-12345678901234567
  anyscale_iam_role_id: arn:aws:iam::123456789012:role/anyscale-iam-role
  cluster_iam_role_id: arn:aws:iam::123456789012:role/cluster-node-role
  memorydb_cluster_name: my-memorydb-cluster
"""

    vpc_id: Optional[str] = field(
        default=None, metadata={"docstring": "The VPC ID."},
    )
    subnet_ids: Optional[List[str]] = field(
        default=None, metadata={"docstring": "List of subnet IDs."},
    )
    zones: Optional[List[str]] = field(
        default=None,
        metadata={
            "docstring": "The availability zone corresponding to each subnet ID."
        },
    )
    security_group_ids: Optional[List[str]] = field(
        default=None, metadata={"docstring": "List of security group IDs."},
    )
    anyscale_iam_role_id: Optional[str] = field(
        default=None, metadata={"docstring": "The Anyscale IAM role ARN."},
    )
    external_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The trust policy external ID for the cross-account IAM role"
        },
    )
    cluster_iam_role_id: Optional[str] = field(
        default=None, metadata={"docstring": "The IAM role ARN used by Ray clusters."},
    )
    memorydb_cluster_name: Optional[str] = field(
        default=None, metadata={"docstring": "The MemoryDB cluster name."},
    )
    memorydb_cluster_arn: Optional[str] = field(
        default=None, metadata={"docstring": "The MemoryDB cluster ARN."},
    )
    memorydb_cluster_endpoint: Optional[str] = field(
        default=None, metadata={"docstring": "The MemoryDB cluster endpoint."},
    )
    cloudformation_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The CloudFormation stack ID, for Anyscale-managed resources."
        },
    )


@dataclass(frozen=True)
class GCPConfig(ModelBase):
    """GCP provider-specific configurations."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
gcp_config:
  project_id: my-project
  provider_name: projects/123456789012/locations/global/workloadIdentityPools/my-cloud/providers/my-provider
  vpc_name: my-vpc
  subnet_names:
    - my-subnet
  firewall_policy_names:
    - my-firewall-policy
  anyscale_service_account_email: my-anyscale-service-account@my-project.iam.gserviceaccount.com
  cluster_service_account_email: my-cluster-service-account@my-project.iam.gserviceaccount.com
  memorystore_instance_name: my-memorystore-instance
"""

    project_id: Optional[str] = field(
        default=None, metadata={"docstring": "The GCP project ID."},
    )
    host_project_id: Optional[str] = field(
        default=None, metadata={"docstring": "The host project ID for shared VPCs."},
    )
    provider_name: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Workload Identity Federation provider name for Anyscale access."
        },
    )
    vpc_name: Optional[str] = field(
        default=None, metadata={"docstring": "VPC name."},
    )
    subnet_names: Optional[List[str]] = field(
        default=None, metadata={"docstring": "List of GCP subnet names."},
    )
    firewall_policy_names: Optional[List[str]] = field(
        default=None, metadata={"docstring": "List of GCP firewall policy names."},
    )
    anyscale_service_account_email: Optional[str] = field(
        default=None, metadata={"docstring": "The Anyscale service account email."},
    )
    cluster_service_account_email: Optional[str] = field(
        default=None,
        metadata={"docstring": "The service account email attached to Ray clusters."},
    )
    memorystore_instance_name: Optional[str] = field(
        default=None, metadata={"docstring": "The Memorystore instance name."},
    )
    memorystore_endpoint: Optional[str] = field(
        default=None, metadata={"docstring": "The Memorystore instance endpoint."},
    )
    deployment_manager_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The deployment manager deployment ID, for Anyscale-managed resources."
        },
    )


@dataclass(frozen=True)
class KubernetesConfig(ModelBase):
    """Kubernetes stack configurations."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
kubernetes_config:
  anyscale_operator_iam_id: arn:aws:iam::123456789012:role/anyscale-operator-role
  zones:
    - us-west-2a
    - us-west-2b
    - us-west-2c
"""

    anyscale_operator_iam_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The cloud provider IAM identity federated with the Anyscale Operator's Kubernetes service account, which will be used by Anyscale control plane for validation during Anyscale Operator bootstrap in the dataplane. IN AWS EKS, this is the ARN of the IAM role. For GCP GKE, this is the service account email."
        },
    )
    zones: Optional[List[str]] = field(
        default=None, metadata={"docstring": "List of zones to launch pods in."},
    )


################################################################################
# NOTE: The CloudResource model below is copied from the OpenAPI CloudDeployment
# model, which is what is actually used in the CLI. It is only defined here so
# that it appears in the generated docs, to provide users with examples of the
# expected YAML format. There is no CloudResource SDK support, so the name of
# this model should not actually matter. (There is also no Cloud SDK support.)
################################################################################
@dataclass(frozen=True)
class CloudResource(ModelBase):
    """Cloud resource configuration."""

    __skip_py_example__ = True

    __doc_yaml_example__ = """\
cloud_resource_id: cldrsrc_12345678901234567890123456
name: my-cloud-resource
provider: AWS
compute_stack: VM
region: us-west-2
networking_mode: PUBLIC
object_storage:
  bucket_name: s3://my-bucket
file_storage:
  file_storage_id: fs-12345678901234567
aws_config:
  vpc_id: vpc-12345678901234567
  subnet_ids:
  - subnet-11111111111111111
  - subnet-22222222222222222
  security_group_ids:
  - sg-12345678901234567
  anyscale_iam_role_id: arn:aws:iam::123456789012:role/anyscale-iam-role
  cluster_iam_role_id: arn:aws:iam::123456789012:role/cluster-node-role
  memorydb_cluster_name: my-memorydb-cluster
"""

    cloud_resource_id: Optional[str] = field(
        default=None,
        metadata={"docstring": "Unique identifier for this cloud resource."},
    )
    name: Optional[str] = field(
        default=None, metadata={"docstring": "The name of this cloud resource."},
    )
    provider: Union[CloudProvider, str] = field(
        default=CloudProvider.UNKNOWN,
        metadata={
            "docstring": "The cloud provider type (e.g., AWS, GCP, AZURE, or GENERIC)."
        },
    )
    compute_stack: Union[ComputeStack, str] = field(
        default=ComputeStack.VM,
        metadata={"docstring": "The compute stack (VM or K8S)."},
    )
    region: Optional[str] = field(
        default=None, metadata={"docstring": "The region (e.g., us-west-2)."},
    )
    networking_mode: Optional[NetworkingMode] = field(
        default=None,
        metadata={"docstring": "Whether to use public or private networking."},
    )
    object_storage: Optional[ObjectStorage] = field(
        default=None, metadata={"docstring": "Object storage configuration."},
    )
    file_storage: Optional[FileStorage] = field(
        default=None, metadata={"docstring": "File storage configuration."},
    )
    aws_config: Optional[AWSConfig] = field(
        default=None, metadata={"docstring": "AWS provider-specific configurations."},
    )
    gcp_config: Optional[GCPConfig] = field(
        default=None, metadata={"docstring": "GCP provider-specific configurations."},
    )
    kubernetes_config: Optional[KubernetesConfig] = field(
        default=None, metadata={"docstring": "Kubernetes stack configurations."},
    )
